# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from collections.abc import Callable

from karabo.common.api import (
    KARABO_SCHEMA_ACCESS_MODE, KARABO_SCHEMA_ASSIGNMENT,
    KARABO_SCHEMA_CLASS_ID, KARABO_SCHEMA_DEFAULT_VALUE, KARABO_SCHEMA_OPTIONS)
from karabo.native import (
    AccessMode, Assignment, Hash, NodeType, Schema, has_changes, is_equal)


def is_init(attrs: dict) -> bool:
    """Check if the passed attributes are related to an init property"""
    mode = attrs[KARABO_SCHEMA_ACCESS_MODE] in {
        AccessMode.RECONFIGURABLE, AccessMode.INITONLY}

    internal = attrs[KARABO_SCHEMA_ASSIGNMENT] == Assignment.INTERNAL
    return mode and not internal


def is_writable(attrs: dict) -> bool:
    """Check if the passed attributes are related to a writable property"""
    mode = attrs[KARABO_SCHEMA_ACCESS_MODE] == AccessMode.RECONFIGURABLE
    return mode


def sanitize_init_configuration(schema: Schema, config: Hash) -> Hash:
    """Sanitize a configuration to be used as INIT configuration

    - Remove all readOnly configuration that are not in the schema!
    - Remove all assignment internal variables

    :param schema: The `Schema` object of the device
    :param config: The configuration `Hash`
    """
    return _extract_configuration(schema, config, False, is_init)


def sanitize_write_configuration(schema: Schema, config: Hash) -> Hash:
    """Sanitize a configuration to be applied as runtime configuration

    :param schema: state dependent runtime schema
    :param config: configuration hash
    """
    return _extract_configuration(schema, config, False, is_writable)


def _extract_configuration(
        schema: Schema, config: Hash, only_changes: bool,
        condition: Callable):
    """Extract a configuration with a reference schema from a config `Hash`

    :param schema: The schema for filtering
    :param config: The configuration to be used

    Note: For runtime use, the schema should be the state dependent schema
    Note: Returns a configuration Hash, that:
        - Does not contain Slots
        - Does not have obsolete paths, e.g. key has to be in schema
        - The configuration won't have `None` values
    """
    assert isinstance(schema, Schema)

    def _iter_schema(schema_hash: Hash, base: str, condition: Callable):
        base = base + '.' if base else ''
        for key, value, attrs in schema_hash.iterall():
            subkey = base + key
            is_slot = attrs.get(KARABO_SCHEMA_CLASS_ID, "") == "Slot"
            is_node = attrs["nodeType"] == NodeType.Node.value
            if is_node and not is_slot:
                yield from _iter_schema(
                    value, base=subkey, condition=condition)
            elif not is_slot and condition(attrs):
                yield subkey, attrs

    retval = Hash()
    for key, attrs in _iter_schema(schema.hash, "", condition):
        value = config.get(key, None)
        # If the key is not in the config, we continue
        if value is None:
            continue
        if only_changes:
            default = attrs.get(KARABO_SCHEMA_DEFAULT_VALUE, None)
            if default is not None and not has_changes(default, value):
                continue
        # Check if we have options and act accordingly, yes
        # some fancy schema evolution can always appear
        options = attrs.get(KARABO_SCHEMA_OPTIONS, None)
        if options is not None and value not in options:
            continue
        retval[key] = value

    return retval


def extract_init_configuration(schema: Schema, config: Hash) -> Hash:
    """Extract a project compatible init only configuration

    - Provides only changes from defaults

    :param schema: The `ClassSchema` object of the device
    :param config: The configuration `Hash`
    """
    return _extract_configuration(schema, config, True, is_init)


def config_changes(a, b):
    """Compare two `Hash` configurations by `key`, `value`

    Note: This function does not consider attributes modifications.

    :param a: input `Hash` a
    :param b: input `Hash` b

    :returns: changes Hash
    """
    assert isinstance(a, Hash), isinstance(b, Hash)

    changes = Hash()
    for key, a_value, _ in Hash.flat_iterall(a):
        if key not in b:
            changes[key] = [a_value, None]
            continue
        b_value = b[key]
        if not is_equal(a_value, b_value):
            changes[key] = [a_value, b_value]

    for key, b_value, _ in Hash.flat_iterall(b):
        if key not in a:
            changes[key] = [None, b_value]

    return changes


def validate_init_configuration(schema: Schema, config: Hash) -> str:
    """Validate an incoming configuration against a schema

    Readonly parameters are removed for the time being. Not existent
    parameters are declared invalid and reported!

    Returns a text (str) which is empty if no violations are found,
    otherwise the issues are reported.
    Hence, an empty string means success!

    :returns: text (str)
    """
    text = ""
    not_available_paths = []

    to_erase = []
    for path, _, attrs in Hash.flat_iterall(config, empty=False):
        try:
            if (schema.hash[path, KARABO_SCHEMA_ACCESS_MODE]
                    == AccessMode.READONLY.value):
                # Note: We can still receive configs for read only params,
                # Extend for read only when attributes are removed.
                to_erase.append(path)
                continue
        except KeyError:
            not_available_paths.append(path)

    if not_available_paths:
        text += ("Configuration contains paths that are not "
                 f"available {not_available_paths}. ")

    to_erase.extend(not_available_paths)
    for path in to_erase:
        config.erase(path)

    # Cleanup for left over nodes (empty Hashes)!
    to_erase = [path for path, v, _ in Hash.flat_iterall(
        config, empty=True) if isinstance(config[path], Hash)
        and config[path].empty()]

    for path in to_erase:
        config.erase(path)

    # Note: If not an empty text is returned, we have invalid paths!
    return text



from datetime import datetime, timezone

from sqlmodel import Field, Relationship, SQLModel, UniqueConstraint


class NamedDeviceInstance(SQLModel, table=True):
    __tablename__ = "NamedDeviceInstance"

    id: int | None = Field(default=None, primary_key=True)
    device_id: str = Field(unique=True, index=True)
    class_id: str = Field(max_length=128, nullable=False)
    server_id: str = Field(max_length=128, nullable=False)

    # One-to-many relationship with NamedDeviceConfig
    configurations: list["NamedDeviceConfig"] = Relationship(
        back_populates="instance", sa_relationship_kwargs={
            "cascade": "all, delete-orphan"})


class NamedDeviceConfig(SQLModel, table=True):
    __tablename__ = "NamedDeviceConfig"

    __table_args__ = (UniqueConstraint(
        "device_id", "name", name="uq_device_name"),)

    id: int | None = Field(default=None, primary_key=True)
    device_id: str = Field(
        foreign_key="NamedDeviceInstance.device_id", index=True)
    name: str
    config_data: str
    date: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc))

    last_loaded: datetime | None = Field(
        default=None, nullable=True)

    # Relationship back to DeviceInstance
    instance: NamedDeviceInstance | None = Relationship(
        back_populates="configurations")



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import time
from asyncio import gather, get_event_loop, sleep
from datetime import datetime

import pytest

from karabo.middlelayer import (
    Device, DeviceClientBase, Hash, KaraboError, Schema, Timestamp, background,
    get_utc_string, getConfigurationFromPast, getInstanceInfo,
    getSchemaFromPast, getTopology, slot, synchronize)
from karabo.middlelayer.testing import AsyncDeviceContext

DEVICE_ID = "data_logger_device_id"

conf = {
    "classId": "DataLogReader",
    "_deviceId_": DEVICE_ID,
}

confd = {
    "classId": "DeviceClient",
    "_deviceId_": "Deviceclient-09",
}


@synchronize
def _getLogReaderId(device):
    return DEVICE_ID


class HiddenDeviceClient(DeviceClientBase):
    def _initInfo(self):
        """Hide this device in the `client` section of the topology

        DeviceClientBase is a device, and this will influence the
        topology that the test `TestDeviceClientBase` expects
        in a non-trivial and time-dependent way.
        """
        info = super()._initInfo()
        info["type"] = "client"
        return info


class DataLogReader(Device):

    def _initInfo(self):
        info = super()._initInfo()
        info["TestInfo"] = "This is a karabo test info"
        return info

    @slot
    async def slotGetConfigurationFromPast(self, deviceId, timepoint):
        if deviceId == "aDeviceNotInHistory":
            raise KaraboError
        h = Hash('value', 42)
        s = Schema(name=DEVICE_ID, hash=Hash())

        return h, s, True, datetime.now().isoformat()


def test_parse_timestamp():
    points = [datetime.now().isoformat(), Timestamp(), None]
    for timepoint in points:
        assert isinstance(get_utc_string(timepoint), str)


@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_getConfigurationSchemaFromPast(mocker):
    async with AsyncDeviceContext(reader=DataLogReader(conf)):
        mocker.patch('karabo.middlelayer.device_client._getLogReaderId',
                     new=_getLogReaderId)
        time = datetime.now().isoformat()

        with pytest.raises(KaraboError):
            await getConfigurationFromPast("aDeviceNotInHistory", time)

        h = await getConfigurationFromPast("aDeviceInHistory", time)
        assert isinstance(h, Hash)
        assert h['value'] == 42

        s = await getSchemaFromPast("aDeviceInHistory", time)
        assert isinstance(s, Schema)
        assert s.name == DEVICE_ID

        info = await getInstanceInfo(DEVICE_ID)
        assert info["TestInfo"] == "This is a karabo test info"
        assert info["classId"] == "DataLogReader"


@pytest.mark.timeout(40)
@pytest.mark.asyncio
async def test_device_client_base():
    dev = HiddenDeviceClient(confd)
    async with AsyncDeviceContext(device=dev):

        class DefaultDevice(Device):
            __version__ = "1.2.3"

        msg = f"Topology has more: {getTopology()['device']}"
        # 1 since signal slotable is there
        assert len(getTopology()["device"]) == 1, msg

        devices = []

        NUMBER_DEVICES = 20
        for number in range(NUMBER_DEVICES):
            dev = DefaultDevice(
                dict(_deviceId_=f"other{number}", serverId="tserver"))
            dev.startInstance()
            devices.append(dev)

        finished = False

        def check_increasing():
            nonlocal finished
            count = 0
            while True:
                num_devices = len(getTopology()["device"])
                count += 1
                if num_devices == NUMBER_DEVICES + 1:
                    # release test
                    finished = True
                    break
                elif count > 25:
                    # Max 25 seconds until a device appears
                    finished = True
                    break
                elif num_devices > 0:
                    # Devices appear in topology, start
                    # checking topology from this thread
                    for _ in range(100):
                        getTopology()
                        time.sleep(0.01)
                else:
                    time.sleep(1)

        # Start a thread in the background
        loop = get_event_loop()
        background(loop.run_coroutine_or_thread(check_increasing))

        # Busy wait until the thread is finished
        while not finished:
            await sleep(0.5)
        assert len(getTopology()["device"]) == NUMBER_DEVICES + 1

        # Kill all devices
        futures = [d.slotKillDevice() for d in devices]
        await gather(*futures, return_exceptions=False)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.native import (
    AccessMode, Assignment, Attribute, KaraboError, String, StringValue,
    get_timestamp, isSet)

from .device_client import getDevice


class MetaProxy:
    def __init__(self, deviceId):
        self.deviceId = StringValue(
            deviceId, timestamp=get_timestamp())
        self.proxy = None

    @property
    def value(self):
        return self.proxy or self.deviceId


class DeviceNode(String):
    """Copy a remote device into the local's namespace

    A device that controls another device may add a device node such that
    it has a proxy to that device while both are running. The controlled
    device may be configured at initialization time.

    As an example, a stage controlling some motor::

        class Stage(Device):
            motor = DeviceNode()

    Then the motor can be acessed under ``self.motor``. Before
    instantiation, a device node just looks like a string, where the
    id of the target device can be entered.
    """
    accessMode = Attribute(AccessMode.INITONLY, dtype=AccessMode)
    assignment = Attribute(Assignment.MANDATORY, dtype=Assignment)
    classId = "deviceNode"

    def toDataAndAttrs(self, value):
        if not isinstance(value, str):
            # If we have a proxy connection, the value is the `Proxy`,
            # otherwise we have a `StringValue`
            value = value.__meta_deviceId
        data, attrs = super().toDataAndAttrs(value)
        return data, attrs

    def checkedInit(self, instance, value=None):
        """Device Nodes cannot have an empty string
        """
        if not value:
            if self.defaultValue:
                return self._initialize(instance, self.defaultValue)
            raise KaraboError('Assignment is mandatory for "{}"'.format(
                self.key))
        return self._initialize(instance, value)

    async def finalize_init(self, instance):
        """Sets a real proxy to the MetaProxy"""
        meta_proxy = instance._getValue(self.key)
        assert isinstance(meta_proxy, MetaProxy)
        assert meta_proxy.proxy is None, "proxy already initialized"

        value = meta_proxy.deviceId
        root = instance.get_root()
        proxy = await root._sigslot.exitStack.enter_async_context(
            getDevice(value))
        proxy.__meta_deviceId = value
        meta_proxy.proxy = proxy

    def _initialize(self, instance, value):
        # This should not happen as we are mandatory, but we never know
        if not isSet(value):
            instance.__dict__[self.key] = None
            return []
        meta_proxy = MetaProxy(value)
        instance.__dict__[self.key] = meta_proxy
        # return an half instantiated deviceNode
        return []

    def __get__(self, instance, owner):
        if instance is None:
            return self

        value = instance._getValue(self.key)
        if isinstance(value, MetaProxy):
            value = value.value
        return value



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy.orm import selectinload
from sqlmodel import SQLModel, insert, select, update
from sqlmodel.ext.asyncio.session import AsyncSession

from .models import NamedDeviceConfig, NamedDeviceInstance
from .utils import (
    ConfigurationDBError, datetime_from_str, datetime_now, datetime_to_str)

_UNDEFINED = "__none__"


class ConfigurationDatabase:
    def __init__(self, db_name: str):
        self.db_name = db_name
        self.engine = create_async_engine(
            f"sqlite+aiosqlite:///{self.db_name}",
            echo=False,
            pool_pre_ping=True,
            connect_args={"timeout": 30})

        self.session = async_sessionmaker(
            bind=self.engine,
            expire_on_commit=False,
            class_=AsyncSession,
            autoflush=False,
            autocommit=False)

    @property
    def path(self) -> Path:
        return Path(self.db_name)

    @property
    def dirname(self) -> str | None:
        return str(self.path.parent) if self.path.exists() else None

    async def assure_existing(self):
        """Ensures the database schema is created."""
        async with self.engine.begin() as conn:
            await conn.run_sync(SQLModel.metadata.create_all)

    async def delete(self):
        """Deletes the database schema and file."""
        async with self.engine.begin() as conn:
            await conn.run_sync(SQLModel.metadata.drop_all)
        await self.engine.dispose()
        if self.path.exists():
            self.path.unlink()

    # Public Interface
    # --------------------------------------------------------------------

    async def list_configurations(
            self, device_id: str, name_part: str = "") -> list[dict]:
        """Retrieves the set of device configurations related to a name part.

        :param device_id: The ID of the device.
        :param name_part: Partial name filter for configurations.
                        If empty, retrieves all configurations.
        :return: List of configuration dictionaries (can be empty).
        """
        async with self.session() as session:
            stmt = (
                select(NamedDeviceInstance)
                .where(NamedDeviceInstance.device_id == device_id)
                # Orm query
                .options(selectinload(NamedDeviceInstance.configurations)))

            result = await session.exec(stmt)
            device = result.first()
            if not device:
                return []

            configurations = device.configurations
            if name_part:
                configurations = [config for config in configurations
                                  if name_part in config.name]

            return [
                {"name": config.name,
                 "serverId": device.server_id,
                 "classId": device.class_id,
                 "date": datetime_to_str(config.date),
                 "last_loaded": datetime_to_str(config.last_loaded)}
                for config in configurations]

    def session(self):
        return self.db_handle

    async def delete_configuration(self, device_id: str, name: str):
        """Delete a device configuration by `name` and `deviceId`"""
        async with self.session() as session:
            stmt = (
                select(NamedDeviceConfig)
                .where(
                    NamedDeviceConfig.device_id == device_id,
                    NamedDeviceConfig.name == name)
            )
            result = await session.exec(stmt)
            config = result.first()
            if not config:
                raise ConfigurationDBError(
                    f"Configuration for {device_id} - {name} not available.")

            try:
                await session.delete(config)
                # Only flush, don't commit, but keep transaction
                await session.flush()
                # Note: Check if the parent device has any remaining
                # configurations
                stmt = (
                    select(NamedDeviceInstance)
                    .where(NamedDeviceInstance.device_id == device_id)
                    .options(selectinload(NamedDeviceInstance.configurations))
                )
                result = await session.exec(stmt)
                device = result.first()
                if device and not device.configurations:
                    await session.delete(device)
                await session.commit()
            except Exception as e:
                await session.rollback()
                raise ConfigurationDBError(
                    f"Failed to delete configuration: {e}")

    async def list_devices(self) -> list[str]:
        """Retrieves the set of deviceIds

        :returns: list of deviceIds records (can be empty).
        """
        async with self.session() as session:
            stmt = select(NamedDeviceInstance.device_id).distinct()
            result = await session.exec(stmt)
            return result.all()

    async def get_configuration(self, device_id: str, name: str) -> dict:
        """Retrieve a device configuration with name and update `last_loaded`

        :param device_id: The ID of the device.
        :param name: The name of the configuration.
        :return: Configuration dictionary (or None if not found).
        """
        stmt = (
            select(NamedDeviceConfig, NamedDeviceInstance)
            .join(
                NamedDeviceInstance,
                NamedDeviceConfig.device_id == NamedDeviceInstance.device_id)
            .where(NamedDeviceConfig.device_id == device_id,
                   NamedDeviceConfig.name == name).with_for_update())
        async with self.session() as session:
            result = await session.exec(stmt)
            row = result.first()
            if not row:
                text = (f"Configuration for device {device_id} "
                        f"with {name} not found.")
                raise ConfigurationDBError(text)
            config, device = row
            update_stmt = (
                update(NamedDeviceConfig)
                .where(NamedDeviceConfig.id == config.id)
                .values(last_loaded=datetime_now()))

            await session.exec(update_stmt)
            await session.commit()

            return {
                "name": name,
                "deviceId": device_id,
                "classId": device.class_id,
                "serverId": device.server_id,
                "date": datetime_to_str(config.date),
                "config": config.config_data}

    async def save_configuration(
            self, name: str, configs: list[dict],
            date: str | None = None):
        """Saves one or more device configurations, ensuring the Device exists.

        :param name: The configuration(s) name.
        :param configs: A list of dictionaries with keys:
                        - "deviceId"
                        - "config"
        :param date: Optional ISO8601 date;
                          defaults to current UTC time.
        """
        if len(name) >= 80:
            text = "Please provide a name with less than 80 characters."
            raise ConfigurationDBError(text)
        required_keys = ["deviceId", "config"]
        for index, config in enumerate(configs):
            if not all(key in config for key in required_keys):
                raise ConfigurationDBError(
                    f"Configuration #{index} is missing required keys.")

        async with self.session() as session:
            try:
                date = (datetime_from_str(date) if date
                        is not None else datetime_now())

                for config in configs:
                    device_id = config["deviceId"]
                    stmt_device = select(NamedDeviceInstance).where(
                        NamedDeviceInstance.device_id == device_id
                        ).with_for_update()
                    device = (await session.exec(
                        stmt_device)).first()
                    class_id = config.get("classId", _UNDEFINED)
                    server_id = config.get("serverId", _UNDEFINED)
                    if not device:
                        session.add(NamedDeviceInstance(
                            device_id=device_id,
                            class_id=class_id,
                            server_id=server_id))
                    else:
                        update_stmt = (
                            update(NamedDeviceInstance)
                            .where(NamedDeviceInstance.device_id == device_id)
                            .values(class_id=class_id, server_id=server_id))
                        await session.exec(update_stmt)
                    stmt_config = (
                        insert(NamedDeviceConfig)
                        .values(
                            device_id=device_id,
                            name=name,
                            date=date,
                            config_data=config["config"]
                        ).prefix_with("OR REPLACE"))
                    await session.exec(stmt_config)
                await session.commit()
            except Exception as e:
                await session.rollback()
                raise ConfigurationDBError(f"Unexpected Error: {e}")



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import datetime
from time import gmtime, strftime, strptime, time

import pytest
from lxml import etree

from karabo.project_db import (
    DATE_FORMAT, ExistDatabase, ProjectDBError, make_xml_if_needed)
from karabo.project_db.exist_db import TESTDB_ADMIN_PASSWORD
from karabo.project_db.testing import (
    create_hierarchy, create_trashed_project, create_unattached_scenes,
    generate_uuid)

from .util import stop_local_database


@pytest.fixture(scope="function")
def database():
    user = "admin"
    password = TESTDB_ADMIN_PASSWORD
    yield ExistDatabase(user, password,
                        test_mode=True, init_db=True)
    stop_local_database()


@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_project_interface(database, subtests):
    # A bunch of document "names" for the following tests
    testproject = generate_uuid()
    testproject2 = generate_uuid()
    async with database as db:
        # remove previously existing test collection
        path = "{}/{}".format(db.root, 'LOCAL_TEST')
        if db.dbhandle.hasCollection(path):
            db.dbhandle.removeCollection(path)

        path = "{}/{}".format(db.root, 'LOCAL')
        if db.dbhandle.hasCollection(path):
            db.dbhandle.removeCollection(path)

        # make sure we have the LOCAL collection and subcollections
        path = "{}/{}".format(db.root, 'LOCAL')
        if not db.dbhandle.hasCollection(path):
            db.dbhandle.createCollection(path)

        with subtests.test(msg='test_domain_exists'):
            assert db.domain_exists("LOCAL")

        with subtests.test(msg='test_add_domain'):
            db.add_domain("LOCAL_TEST")
            assert db.domain_exists("LOCAL_TEST")

        with subtests.test(msg='test_save_item'):
            # The MySQL back-end doesn't accept project items with no
            # name or no type - originally this test's xml only had 'uuid'.
            xml_rep = (
                f'<xml uuid="{testproject2}" item_type="device_server" '
                'simple_name="xyz">foo</xml>')

            date = await db.save_item('LOCAL', testproject2, xml_rep)
            path = f"{db.root}/LOCAL/{testproject2}_0"
            assert db.dbhandle.hasDocument(path)
            decoded = db.dbhandle.getDoc(path).decode('utf-8')
            doctree = etree.fromstring(decoded)
            assert doctree.get('uuid') == testproject2
            assert doctree.text == 'foo'
            assert isinstance(date, str)

        with subtests.test(msg='test_save_bad_item'):
            xml_rep = """
            <xml uuid="{uuid}"><foo|bar/></xml>
            """.format(uuid=testproject)
            with pytest.raises(ProjectDBError):
                await db.save_item('LOCAL', testproject, xml_rep)

        with subtests.test(msg='load_items'):
            items = [{"uuid": testproject2, "item_type": "nocare"},
                     {"uuid": testproject, "item_type": "nocare"}]
            item_uuids = [it["uuid"] for it in items]

            res = await db.load_item('LOCAL', items)
            for r in res:
                itemxml = make_xml_if_needed(r["xml"])
                assert itemxml.tag == 'xml'
                # The MySQL back-end does not save the inner "foo" text of
                # the original XML as it doesn't map to any item field and
                # is ignored. The previous assert for itemxml.text == 'foo'
                # has thus been replaced by a UUID check
                assert itemxml.attrib['uuid'] in item_uuids

        with subtests.test(msg='test_list_items'):
            await create_hierarchy(db)
            items = await db.list_items('LOCAL', ['project', 'scene'])
            assert len(items) == 5
            scenecnt = 0
            for i in items:
                if i["item_type"] == "scene":
                    scenecnt += 1
                # The contract of "list_items" is to return a
                # list of items (dictionaries) where
                # each item would have keys "uuid", "item_type"
                # and "simple_name" (no mention to "is_trashed")
                if "is_trashed" in i.keys():
                    assert i["is_trashed"] is False
            assert scenecnt >= 4

        with subtests.test(msg='test_trashed_projects'):
            await create_trashed_project(db)
            items = await db.list_items('LOCAL', ['project'])
            # Before the assertion was for number of projects being 2, but
            # for the MySQL not all the save operations are invoked.
            assert len(items) >= 1
            nb_is_trashed = 0
            for it in items:
                if it["is_trashed"] is True:
                    nb_is_trashed += 1
            assert nb_is_trashed == 1

        with subtests.test(msg='test_list_domains'):
            items = await db.list_domains()
            for it in items:
                assert it in ['LOCAL_TEST', 'REPO', 'LOCAL']

        with subtests.test(msg='test_old_versioned_data'):
            versioned_uuid = generate_uuid()
            xml_rep = """
            <xml uuid="{uuid}" revision="{revision}" date="{date}"
                user="sue">foo</xml>
            """
            MAX_REV = 3
            earlier = time() - 5.0
            # Save three consecutive versions
            for i in range(MAX_REV, 0, -1):
                # consecutive dates beginning a few seconds ago
                date = strftime(DATE_FORMAT, gmtime(earlier + i))
                xml = xml_rep.format(uuid=versioned_uuid, revision=i,
                                     date=date)
                path = "{}/{}/{}_{}".format(
                    db.root, 'LOCAL', versioned_uuid, i)
                # Save directly without engaging save_item()
                db.dbhandle.load(xml.encode('utf-8'), path)

            # Read it back, no revision given
            items = [{"uuid": versioned_uuid, "item_type": "notimportant"}]
            res = await db.load_item('LOCAL', items)
            assert len(res) == 1
            itemxml = make_xml_if_needed(res[0]['xml'])
            assert int(itemxml.get('revision')) == MAX_REV

            # Write it back with save_item
            date = strftime(DATE_FORMAT, gmtime())
            xml_rep = """
            <xml uuid="{uuid}" date="{date}" user="sue">foo</xml>
            """.format(uuid=versioned_uuid, date=date)
            await db.save_item('LOCAL', versioned_uuid, xml_rep)

            # Make sure there's a revision of 0 (auto-added)
            items = [{"uuid": versioned_uuid, "item_type": "notimportant"}]
            res = await db.load_item('LOCAL', items)
            itemxml = make_xml_if_needed(res[0]['xml'])
            assert itemxml.get('revision') == '0'

        with subtests.test(msg='test_named_items'):
            await create_unattached_scenes(db)
            items = await db.list_named_items(
                'LOCAL', 'scene', 'Scene!')
            assert len(items) == 4
            scenecnt = 0
            for i in items:
                if i["item_type"] == "scene":
                    scenecnt += 1
                assert i["simple_name"] == "Scene!"
            assert scenecnt >= 4

        stop_local_database()


@pytest.mark.timeout(30)
@pytest.mark.asyncio
async def test_save_check_modification(database, subtests):
    proj_uuid = generate_uuid()

    async with database as db:
        path = "{}/{}".format(db.root, 'LOCAL')
        if db.dbhandle.hasCollection(path):
            db.dbhandle.removeCollection(path)

        # make sure we have the LOCAL collection and subcollections
        path = "{}/{}".format(db.root, 'LOCAL')
        if not db.dbhandle.hasCollection(path):
            db.dbhandle.createCollection(path)

        with subtests.test(msg='test_save_item'):
            date = "2011-11-01 09:00:52"
            # The MySQL back-end doesn't accept project items with no
            # name or no type - originally this test's xml only had 'uuid'
            # and 'date'
            xml_rep = """
            <xml uuid="{uuid}" date="{date}" simple_name="a_scene"
            item_type="scene">foo</xml>
            """.format(uuid=proj_uuid, date=date)

            success = True
            try:
                new_date = await db.save_item('LOCAL', proj_uuid, xml_rep)
            except ProjectDBError:
                success = False

            assert success
            assert new_date != date
            date_t = strptime(date, DATE_FORMAT)
            try:
                current_date_t = strptime(new_date, DATE_FORMAT)
            except ValueError:
                # The MySQL back-end sends an UTC date in ISO8601 format
                # in the metadata returned by save_item.
                # Try to parse the date using that format.
                # TODO: FIX
                current_date_t = (
                    datetime.datetime.fromisoformat(new_date).timetuple())

            assert date_t < current_date_t

            # Try saving again but using same time stamp
            success = True
            reason = ""
            try:
                await db.save_item('LOCAL', proj_uuid, xml_rep)
            except ProjectDBError as e:
                success = False
                reason = str(e)
            assert not success
            assert reason



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import argparse

from traits.api import Instance, List

from karabo.common.api import BaseSavableModel
from karabo.common.scenemodel.api import FixedLayoutChildData, read_scene


def code_for(
        model: BaseSavableModel,
        children_keys: list[str],
        children: list[list[str]] | None = None) -> tuple[set[str], list[str]]:
    """Generate the Python code which can build a scene model object.

    :param children_keys: the list of children item key, in the model.
    """
    ignored_traits = (*children_keys, 'uuid')
    traits = []
    for name in sorted(model.copyable_trait_names()):
        value = getattr(model, name)
        if name in ignored_traits or isinstance(value, FixedLayoutChildData):
            continue

        trait_default = model.trait(name).default
        if value == trait_default:
            continue

        traits.append(f'{name}={repr(value)}')

    if children:
        for key, child in zip(children_keys, children):
            if child:
                traits.append('{}=[{}]'.format(key, ', '.join(child)))

    klass_name = model.__class__.__name__
    return klass_name, '{}({})'.format(klass_name, ', '.join(traits))


def _list_of_savable(trait):
    if not isinstance(trait.trait_type, List):
        return False
    inner_type = trait.inner_traits[0].trait_type
    if not isinstance(inner_type, Instance):
        return False
    if not issubclass(inner_type.klass, BaseSavableModel):
        return False
    return True


def _find_children(obj):
    children = [name for name in obj.copyable_trait_names()
                if _list_of_savable(obj.trait(name))]
    if not children:
        children.append("children")
    return children


def convert_scene_model_to_code(
        model: BaseSavableModel,
        name: str = 'scene') -> tuple[set[str], list[str]]:
    """Given a `SceneModel` object, generate the Python code which can
    reproduce it.
    """
    code = []
    classes = set()

    # Note: Compatibility, not all models have a `children` list, but others.
    # If we find multiple children, we still choose `children`
    keys = _find_children(model)

    child_names = []
    for j, key in enumerate(keys):
        children = getattr(model, key, [])
        key_child_names = []
        for i, child in enumerate(children):
            child_name = name + str(j) + str(i)
            key_child_names.append(child_name)
            child_cls, child_stmts = convert_scene_model_to_code(
                child, name=child_name)
            code.extend(child_stmts)
            classes.update(child_cls)
        child_names.append(key_child_names)

    klass_name, stmt = code_for(model, keys, children=child_names)
    code.append(f'{name} = {stmt}')
    classes.add(klass_name)
    return classes, code


def main():
    description = 'Convert a Karabo scene file to Python code.'
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('filename',
                        help='An SVG file containing a Karabo scene')
    parser.add_argument('deviceId', nargs='?', default=None,
                        help='deviceId of the source device'
                             ' (optional argument)')
    ns = parser.parse_args()

    # main scene code generation
    symbols, code = convert_scene_model_to_code(read_scene(ns.filename))

    # import statement
    symbols.update({'write_scene', 'SceneTargetWindow'})
    imp_stmt = ('from karabo.common.scenemodel.api import {}'
                ''.format(', '.join(sorted(symbols))))
    code.insert(0, imp_stmt)

    # Dump it out
    indent = '\n' + ' ' * 4
    output_scene = code[0]  # <-- import statement
    output_scene += '\n\ndef get_scene(deviceId):\n'
    output_scene += ' ' * 4 + indent.join(code[1:])
    output_scene += indent + 'return write_scene(scene)\n'

    # change Ubuntu font (can be removed after some transitional period)
    output_scene = output_scene.replace("font='Ubuntu,",
                                        "font='Source Sans Pro,")

    # Substitute deviceId and print
    if ns.deviceId:
        if not output_scene.count(f"keys=['{ns.deviceId}"):
            print(f"# WARNING from scene2py: "
                  f"no occurrences of '{ns.deviceId}' found")

        output = output_scene.replace(f"keys=['{ns.deviceId}",
                                      "keys=[f'{deviceId}")
        output = output.replace(f"'{ns.deviceId}", "f'{deviceId}")
        output = output.replace(f"{ns.deviceId}", "{deviceId}")
        print(output)
    else:
        print(output_scene)


if __name__ == '__main__':
    main()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from asyncio import TimeoutError, get_event_loop, sleep

import numpy as np
import pytest
import pytest_asyncio

from karabo import __version__ as karaboVersion
from karabo.common.states import State
from karabo.middlelayer import (
    AccessMode, KaraboError, background, getDevice, getOutputChannelInfo,
    setWait, sleep as mdlsleep, waitUntil, waitWhile)
from karabo.middlelayer.device import Device
from karabo.middlelayer.device_client import (
    call, get_instance, getProperties, getSchema)
from karabo.middlelayer.pipeline import InputChannel, OutputChannel
from karabo.middlelayer.signalslot import slot
from karabo.middlelayer.testing import AsyncDeviceContext, run_test, sleepUntil
from karabo.middlelayer.utils import AsyncTimer, get_property, set_property
from karabo.native import (
    Bool, Configurable, Float, Hash, Int32, Node, Slot, Timestamp, VectorHash)


class RowSchema(Configurable):
    x = Float(defaultValue=1.0)
    y = Float(defaultValue=1.0)


class ChannelData(Configurable):
    """The properties for the pipeline channels"""
    floatProperty = Float(defaultValue=10.0,
                          displayedName="Float")


class PropertyData(Configurable):
    floatProperty = Float(defaultValue=10.0,
                          displayedName="Float",
                          allowedStates=[State.ON])
    intProperty = Int32(defaultValue=7,
                        displayedName="Integer",
                        allowedStates=[State.ERROR])


class ChannelNode(Configurable):
    """A node with an output channel"""
    output = OutputChannel()


class NodeWithSlot(Configurable):
    """A slot in a node"""
    @Slot()
    async def pressMe(self):
        return 6


class MyDevice(Device):
    __version__ = "1.2.3"

    preInitError = Bool(
        defaultValue=False,
        description="Decide if this device's preInitialization should raise")

    initError = Bool(
        defaultValue=False,
        description="Decide if this device should throw an error in "
                    "onInitialization")

    isDown = Bool(
        defaultValue=False,
        description="Indicator if this device went down. This is set"
                    "on slotKillDevice")

    integer = Int32(
        defaultValue=0,
        minInc=0,
        maxInc=10)

    counter = Int32(
        defaultValue=0,
        minInc=0,
        maxInc=2000)

    # an output channel without schema
    output = OutputChannel()
    # output channel with schema
    dataOutput = OutputChannel(ChannelData)
    # output channel in a node
    nodeOutput = Node(ChannelNode)

    noded = Node(PropertyData)

    @InputChannel()
    async def input(self, data, meta):
        """The input channel with data and meta
        """

    table = VectorHash(
        rows=RowSchema,
        displayedName="Table",
        defaultValue=[Hash("x", 2.0, "y", 5.6),
                      Hash("x", 1.0, "y", 1.6)],
        accessMode=AccessMode.RECONFIGURABLE)

    @Slot(displayedName="Increase", allowedStates=[State.ON])
    async def increaseCounter(self):
        self.state = State.ACQUIRING
        self.counter = self.counter.value + 1
        background(self._state_sleep)

    async def _state_sleep(self):
        await sleep(0.2)
        self.state = State.ON

    async def set_state(self, state):
        self.state = state

    @Slot(displayedName="start")
    async def start(self):
        return 5

    async def preInitialization(self):
        if self.preInitError:
            raise RuntimeError("I am going down")

    async def onInitialization(self):
        self.state = State.ON
        if self.initError:
            raise RuntimeError("Status must not be empty")

    async def slotKillDevice(self):
        self.isDown = True
        return await super().slotKillDevice()

    async def setState(self, state):
        self.state = state

    nodeWithSlot = Node(NodeWithSlot)


@pytest_asyncio.fixture(scope="module")
@pytest.mark.asyncio
async def deviceTest():
    myDevice = MyDevice(dict(_deviceId_="MyDevice"))
    async with AsyncDeviceContext(myDevice=myDevice) as ctx:
        yield ctx


@pytest.mark.timeout(30)
@run_test
def test_device_version(deviceTest):
    myDevice = deviceTest["myDevice"]
    expected = "karabo-1.2.3"
    assert myDevice.classVersion == expected
    assert myDevice.karaboVersion == karaboVersion


@pytest.mark.timeout(30)
@run_test
def test_output_names(deviceTest):
    myDevice = deviceTest["myDevice"]
    names = myDevice.slotGetOutputChannelNames()
    expected = ["dataOutput", "nodeOutput.output", "output"]
    assert names == expected


@pytest.mark.timeout(30)
@run_test
def test_displayType_state(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.state.descriptor.displayType == "State"


@pytest.mark.timeout(30)
@run_test
def test_displayType_output(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.output.displayType == "OutputChannel"
    assert myDevice.dataOutput.displayType == "OutputChannel"


@pytest.mark.timeout(30)
@run_test
def test_displayType_input(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.input.displayType, "InputChannel"


@pytest.mark.timeout(30)
@run_test
def test_displayType_lockedBy(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.lockedBy.descriptor.displayType == "lockedBy"


@pytest.mark.timeout(30)
@run_test
def test_classId_output(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.output.classId == "OutputChannel"
    assert myDevice.dataOutput.classId == "OutputChannel"


@pytest.mark.timeout(30)
@run_test
def test_classId_input(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.input.classId == "InputChannel"


@pytest.mark.timeout(30)
@run_test
async def test_state_dependent_schema(deviceTest):
    myDevice = deviceTest["myDevice"]
    schema_before = await getSchema("MyDevice", onlyCurrentState=False)
    # The notation `before` refers to the device in state ON.
    with (await getDevice("MyDevice")) as d:
        # Set a transient state
        await myDevice.set_state(State.ACQUIRING)
        proxy_before_schema = await getSchema(d, onlyCurrentState=False)
        assert schema_before.hash.paths() == proxy_before_schema.hash.paths()
        schema_after = await getSchema("MyDevice", onlyCurrentState=True)
        p_after_schema = await getSchema(d, onlyCurrentState=True)
        # Back to State.ON
        await myDevice.set_state(State.ON)
        assert proxy_before_schema.hash.paths() != p_after_schema.hash.paths()
        assert schema_before.hash.paths() != schema_after.hash.paths()
        # Check slot in schemas before (ON) and after (ACQUIRING)
        assert "increaseCounter" in schema_before.hash
        assert "increaseCounter" not in schema_after.hash
        assert "increaseCounter" in proxy_before_schema.hash
        assert "increaseCounter" not in p_after_schema.hash


@pytest.mark.timeout(30)
@run_test
async def test_classId_slot(deviceTest):
    s = await getSchema("MyDevice")
    assert s.hash.hasAttribute("start", "classId")
    assert s.hash.getAttribute("start", "classId") == "Slot"


@pytest.mark.timeout(30)
@run_test
async def test_classId_node(deviceTest):
    s = await getSchema("MyDevice")
    assert not s.hash.hasAttribute("nodeOutput", "classId")


@pytest.mark.timeout(30)
@run_test
async def test_send_raw(deviceTest):
    myDevice = deviceTest["myDevice"]
    hsh = Hash("Itchy", 10)
    assert myDevice.output.schema is None
    await myDevice.output.writeRawData(hsh)

    # provoke attribute error because we don"t have a schema
    with pytest.raises(AttributeError):
        await myDevice.output.writeData(hsh)


@pytest.mark.timeout(30)
@run_test
def test_send_raw_no_wait(deviceTest):
    myDevice = deviceTest["myDevice"]
    hsh = Hash("Scratchy", 20)
    assert myDevice.output.schema is None
    myDevice.output.writeRawDataNoWait(hsh)

    # provoke attribute error because we don"t have a schema
    with pytest.raises(AttributeError):
        myDevice.output.writeDataNoWait(hsh)


@pytest.mark.timeout(30)
@run_test
async def test_lastCommand(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.lastCommand == ""
    leadId = get_event_loop().instance().deviceId
    d = await getDevice("MyDevice")
    with d:
        await d.nodeWithSlot.pressMe()
        assert myDevice.lastCommand == f"nodeWithSlot.pressMe <- {leadId}"
        await d.start()
    assert myDevice.lastCommand == f"start <- {leadId}"
    await getSchema("MyDevice")
    assert myDevice.lastCommand == f"start <- {leadId}"
    with d:
        current = d.integer
        await setWait(d, integer=2)
        assert d.lastCommand == f"slotReconfigure <- {leadId}"
        await setWait(d, integer=current)


@pytest.mark.timeout(30)
@run_test
async def test_two_calls_concurrent(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.counter == 0
    with (await getDevice("MyDevice")) as d:
        await d.increaseCounter()
        await waitUntil(lambda: d.state != State.ON)
        await waitWhile(lambda: d.state == State.ACQUIRING)
        assert myDevice.counter == 1
        # Concurrent slot calls, one will return due to state block
        myDevice.callNoWait("MyDevice", "increaseCounter")
        myDevice.callNoWait("MyDevice", "increaseCounter")
        await waitUntil(lambda: d.state != State.ON)
        await waitWhile(lambda: d.state == State.ACQUIRING)
        assert myDevice.counter == 2


@pytest.mark.timeout(30)
@run_test
async def test_clear_table_external(deviceTest):
    with (await getDevice("MyDevice")) as d:
        dtype = d.table.descriptor.dtype
        current_value = np.array((2.0, 5.6), dtype=dtype)
        assert d.table[0].value == current_value

        # pop the first item and compare
        check_value = d.table.pop(0)
        assert check_value.value == current_value

        # The former second is now first
        current_value = np.array((1.0, 1.6), dtype=dtype)
        assert d.table[0].value == current_value
        assert len(d.table.value) == 1

        # clear the value on the device side. The values are popped.
        d.table.clear()
        empty_table = np.array([], dtype=dtype)
        success = np.array_equal(d.table.value, empty_table)
        assert success


@pytest.mark.timeout(30)
@run_test
async def test_allowed_state_reconfigure_nodes(deviceTest):
    with (await getDevice("MyDevice")) as d:
        assert d.noded.floatProperty == 10.0
        await setWait(d, "noded.floatProperty", 27.0)
        assert d.noded.floatProperty == 27.0
        await setWait(d, "noded.floatProperty", 10.0)
        assert d.noded.floatProperty == 10.0
        with pytest.raises(KaraboError):
            await setWait(d, "noded.intProperty", 22)
        assert d.noded.intProperty == 7


@pytest.mark.timeout(30)
@run_test
def test_slot_verification(deviceTest):
    myDevice = deviceTest["myDevice"]
    assert myDevice.slotHasSlot("increaseCounter") is True
    assert myDevice.slotHasSlot("output") is False
    assert myDevice.slotHasSlot("doesNotExist") is False


@pytest.mark.timeout(30)
@run_test
async def test_instance_info(deviceTest):
    device = deviceTest["myDevice"]
    assert device._sigslot.info["status"] == "ok"
    await device.setState(State.ERROR)
    await sleep(0)
    assert device._sigslot.info["status"] == "error"
    await device.setState(State.UNKNOWN)
    await sleep(0)
    assert device._sigslot.info["status"] == "unknown"
    await device.setState(State.ON)
    await sleep(0)
    assert device._sigslot.info["status"] == "ok"


@pytest.mark.timeout(30)
@run_test
async def test_output_information(deviceTest):
    device = deviceTest["myDevice"]
    # Second argument processId is not used in MDL
    h = device.slotGetOutputChannelInformation(
        Hash("channelId", "output"))
    assert h["success"]
    data = h["info"]
    assert data["connectionType"] == "tcp"
    assert data["memoryLocation"] == "remote"
    assert isinstance(data["port"], np.uint32)

    info = await getOutputChannelInfo(device.deviceId, "output")
    info = info["info"]
    assert info["connectionType"] == "tcp"
    assert info["memoryLocation"] == "remote"
    assert isinstance(info["port"], np.uint32)
    assert info["port"] == data["port"]

    h = device.slotGetOutputChannelInformation(
        Hash("channelId", "doesNotExist"))
    assert not h["success"]
    assert h["info"] == Hash()


@pytest.mark.timeout(30)
@run_test
async def test_get_property_set_property(deviceTest):
    myDevice = deviceTest["myDevice"]

    prop = get_property(myDevice, "integer")
    set_property(myDevice, "integer", 0)
    assert prop == 0
    set_property(myDevice, "noded.floatProperty", 10)
    prop = get_property(myDevice, "noded.floatProperty")
    assert prop == 10
    set_property(myDevice, "noded.floatProperty", 15)
    assert prop is not myDevice.noded.floatProperty
    prop = get_property(myDevice, "noded.floatProperty")
    assert prop == 15
    with pytest.raises(AttributeError):
        get_property(myDevice, "noded.notthere")
    with pytest.raises(AttributeError):
        get_property(myDevice, "notthere")
    with pytest.raises(AttributeError):
        set_property(myDevice, "notthere", 25)
    with pytest.raises(AttributeError):
        set_property(myDevice, "noded.notthere", 2)


@pytest.mark.timeout(30)
@run_test
async def test_get_properties_hash(deviceTest):
    myDevice = deviceTest["myDevice"]
    myDevice.integer = 0

    # Remote slot call via slotGetConfigurationSlice
    deviceId = myDevice.deviceId
    h = await getProperties(deviceId, "integer")
    assert isinstance(h, Hash)
    assert h["integer"] == 0
    assert len(h.paths(intermediate=False)) == 1

    slot = "slotGetConfigurationSlice"
    r = await get_instance().call(
        deviceId, slot, Hash("paths", ["integer"]))
    assert r.fullyEqual(h)

    # And we have the timestamp
    attrs = h["integer", ...]
    assert len(attrs) == 3
    assert "tid" in attrs
    assert "frac" in attrs
    assert "sec" in attrs

    set_property(myDevice, "noded.floatProperty", 15)
    h = await getProperties(deviceId,
                            ["integer", "noded.floatProperty"])
    assert h["integer"] == 0
    assert h["noded.floatProperty"] == 15
    assert len(h.paths(intermediate=True)) == 3
    assert len(h.paths(intermediate=False)) == 2

    # exception for failing properties
    with pytest.raises(KaraboError):
        h = await getProperties(deviceId, ["integer", "noded.notthere"])

    # KaraboError for failing garbage
    for fail in (
        [False, True],
        [Hash()],
        [1, 2],
        [1.78],
    ):
        with pytest.raises(KaraboError):
            h = await getProperties(deviceId, fail)

    # Working with proxy
    async with getDevice(myDevice.deviceId) as proxy:
        h = await getProperties(proxy, ["integer", "noded.floatProperty"])
        assert h["integer"] == 0
        assert h["noded.floatProperty"] == 15
        assert len(h.paths(intermediate=False)) == 2


@pytest.mark.timeout(30)
@run_test
async def test_slot_time(deviceTest):
    h = await call("MyDevice", "slotGetTime")
    assert h is not None
    assert h["time"]
    assert h.getAttributes("time")["sec"] > 0
    assert h.getAttributes("time")["frac"] > 0
    assert h.getAttributes("time")["tid"] == 0
    timestamp_first = Timestamp.fromHashAttributes(h["time", ...])
    h = await call("MyDevice", "slotGetTime")
    timestamp_second = Timestamp.fromHashAttributes(h["time", ...])
    assert timestamp_second > timestamp_first


@pytest.mark.timeout(30)
@run_test
async def test_initialization(deviceTest):
    # this will not throw in the preInitialization
    bob = MyDevice({"_deviceId_": "bob", "preInitError": True})
    # fails in preInitialization -> instantiation will throw
    with pytest.raises(RuntimeError):
        await bob.startInstance()
    assert bob.isDown
    assert not bob.is_initialized

    # this will not throw in the onInitialization
    alice = MyDevice({"_deviceId_": "alice", "initError": True})
    await alice.startInstance()
    # for backward compatibility, we kill the device on initError.
    assert alice.isDown
    assert not alice.is_initialized

    # this will succeed
    charlie = MyDevice({"_deviceId_": "charlie"})
    await charlie.startInstance()
    assert not charlie.isDown
    assert charlie.is_initialized
    await charlie.slotKillDevice()


@pytest.mark.timeout(30)
@run_test
async def test_preinitialization(deviceTest):
    """Test the preinitialization taking too long"""
    class ASlowInitDevice(Device):

        async def preInitialization(self):
            await sleep(10)

        async def slotKillDevice(self):
            self.isDown = True
            await super().slotKillDevice()

    charlie = ASlowInitDevice({"_deviceId_": "charlie"})
    with pytest.raises(TimeoutError):
        await charlie.startInstance()
    assert charlie.isDown
    assert not charlie.is_initialized

    class SlowInitDevice(Device):

        def preInitialization(self):
            mdlsleep(10)

        async def slotKillDevice(self):
            self.isDown = True
            await super().slotKillDevice()

    echo = SlowInitDevice({"_deviceId_": "echo"})
    with pytest.raises(TimeoutError):
        await echo.startInstance()
    assert echo.isDown
    assert not echo.is_initialized


@pytest.mark.timeout(30)
@run_test
async def test_atimer_destruct(deviceTest):

    global counter
    counter = 0

    class TimerDevice(Device):
        __version__ = "1.2.3"
        killed = False

        async def onInitialization(self):
            self.timer = AsyncTimer(self.timer_callback, 0.1)
            self.timer.start()
            get_event_loop().something_changed()

        async def timer_callback(self):
            global counter
            counter += 1

        async def slotKillDevice(self, message=None):
            ret = await super().slotKillDevice(message)
            self.killed = True
            return ret

        slotKillDevice = slot(slotKillDevice, passMessage=True)

    device = TimerDevice({"_deviceId_": "timerDeviceTest"})
    await device.startInstance()
    assert counter == 0
    await waitUntil(lambda: device.is_initialized)
    await sleep(0.2)
    assert counter > 0
    await device.slotKillDevice()
    old_counter = counter
    await sleep(0.2)
    assert counter == old_counter

    counter = 0
    device = TimerDevice({"_deviceId_": "timerDeviceTest"})
    await device.startInstance()
    await waitUntil(lambda: device.is_initialized)
    await sleep(0.2)
    assert counter > 0
    device.__del__()
    await sleepUntil(lambda: device.killed, 5)
    old_counter = counter
    await sleep(0.2)
    assert counter == old_counter

    counter = 0
    device = TimerDevice({"_deviceId_": "timerDeviceTest"})
    await device.startInstance()
    await waitUntil(lambda: device.is_initialized)
    await device.slotKillDevice()
    # Between start and shutdown there might have been a tick,
    # but the timer gracefully stopped
    assert counter <= 1
    assert not device.timer.is_running()



import os
import shutil
import subprocess
from asyncio import (
    create_subprocess_exec, ensure_future, get_event_loop, sleep, wait_for)
from contextlib import suppress
from typing import Any


class AsyncServerContext:
    STARTUP_TIME: int = 1
    WAIT_TIME: int = 10

    def __init__(self, serverId: str, args: list[str],
                 api: str = "middlelayer", verbose: bool = True):
        assert "/" not in serverId, "Test servers can't have `/` ..."
        if args is None:
            args = []
        self.process: subprocess.Popen[Any] = None
        self.args: list[str] = [f"serverId={serverId}"] + args
        if api == "mdl":
            api = "middlelayer"
        elif api == "bound":
            api = "python"
        self.serverId: str = serverId
        self.api: str = api
        self.verbose: bool = verbose

    async def is_alive(self) -> bool:
        if self.process is None:
            return False
        with suppress(TimeoutError):
            await wait_for(self.process.wait(), 0.5)
        return self.process.returncode is None

    async def __aenter__(self) -> "AsyncServerContext":
        process = await create_subprocess_exec(
            f"karabo-{self.api}server", *self.args,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.process = process
        if process is not None:

            async def print_stdout():
                while not process.stdout.at_eof():
                    line = await process.stdout.readline()
                    print(line.decode("ascii"), end="")

            async def print_stderr():
                while not process.stderr.at_eof():
                    line = await process.stderr.readline()
                    print(line.decode("ascii"), end="")

            if self.verbose:
                ensure_future(print_stdout())
                ensure_future(print_stderr())

            await sleep(self.STARTUP_TIME)
        alive = await self.is_alive()
        if not alive:
            raise RuntimeError(f"Could not start {self.serverId}")
        await self.waitOnline(self.WAIT_TIME)
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        if self.process is not None:
            self.process.terminate()
            try:
                await wait_for(self.process.wait(), timeout=10)
            except TimeoutError:
                self.process.kill()

        # Always check for logs
        env = os.getenv("KARABO")
        log_path = f"{env}/var/log/{self.serverId}"
        if os.path.exists(log_path):
            shutil.rmtree(log_path)

    async def waitOnline(self, timeout=10):
        loop = get_event_loop()
        instance = loop.instance()
        start = loop.time()
        while True:
            try:
                await wait_for(instance.call(
                    self.serverId, "slotPing", 1),
                    timeout=0.2)
                break
            except TimeoutError:
                now = loop.time()
                if now - start > timeout:
                    raise TimeoutError(
                        f"Timeout for server {self.serverId}")
                await sleep(0.5)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import base64
import logging
from asyncio import gather
from collections.abc import Awaitable, Callable
from datetime import UTC
from pathlib import Path

from lxml import etree
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import selectinload
from sqlmodel import SQLModel, select

from karabo.common.project.api import PROJECT_DB_SCHEMA

from ..bases import DatabaseBase
from ..util import ProjectDBError, make_str_if_needed, make_xml_if_needed
from .db_engine import create_local_engine, create_remote_engine
from .models import (
    DatabaseMetadata, DeviceConfig, DeviceInstance, DeviceServer, Macro,
    Project, ProjectDomain, ProjectSubproject, Scene, SceneLinkedScene)
from .models_xml import (
    emit_config_xml, emit_device_xml, emit_macro_xml, emit_project_xml,
    emit_scene_xml, emit_server_xml)
from .utils import (
    date_utc_to_local, datetime_from_str, datetime_now, datetime_str_now,
    datetime_to_str, get_scene_links, get_trashed)

logger = logging.getLogger(__file__)


class SQLDatabase(DatabaseBase):

    def __init__(self, user: str = "", password: str = "",
                 server: str = "", port: int = -1, db_name: str = "local.db",
                 local: bool = False):
        super().__init__()
        if local:
            self.db_engine, self.session_gen = create_local_engine(db_name)
        else:
            self.db_engine, self.session_gen = (
                create_remote_engine(user, password, server, port, db_name))
        self.dbName = db_name
        self.initialized = False
        self.local = local

    async def initialize(self):
        async with self.db_engine.begin() as conn:
            await conn.run_sync(SQLModel.metadata.create_all)

        self.initialized = True

    async def assure_domains(self, required_domains: list[str]):
        """Makes sure that the specified required_domains are present in the
           database."""
        if not required_domains:
            raise RuntimeError("'required_domains' cannot be an empty list")
        for domain in required_domains:
            await self.add_domain(domain.strip())

    async def delete(self):
        """Deletes the database schema and file."""
        async with self.db_engine.begin() as conn:
            await conn.run_sync(SQLModel.metadata.drop_all)
        await self.db_engine.dispose()
        if self.local:
            path = Path(self.dbName)
            if path.is_file():
                path.unlink()

    async def __aenter__(self):
        if not self.initialized:
            raise RuntimeError("Database needs to be initialized")
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        return await super().__aexit__(exc_type, exc_value, traceback)

    async def domain_exists(self, domain: str) -> bool:
        domains = await self.list_domains()
        return domain in domains

    async def schema_version(self) -> int:
        metadata = await self._execute_first(select(DatabaseMetadata))
        if metadata is None:
            return PROJECT_DB_SCHEMA
        else:
            return metadata.schema_version

    async def list_items(
            self, domain: str, item_types: list[str] | None = None
    ) -> list[dict[str, any]]:
        """
        List items in domain which match item_types if given, or all items
        if not given.
        :param domain: domain to list items from
        :param item_types: list or tuple of item_types to list
        :return: list of dicts with keys: uuid, item_type, simple_name
        """
        if not await self.domain_exists(domain):
            raise ProjectDBError(f'Domain "{domain}" not found')

        if item_types is None:
            item_types = [
                "project", "macro", "scene",
                "device_server", "device_instance", "device_config"
            ]

        reader_map: dict[str, Callable[[str], Awaitable[list]]] = {
            "project": self._get_domain_projects,
            "macro": self._get_domain_macros,
            "scene": self._get_domain_scenes,
            "device_server": self._get_domain_device_servers,
            "device_instance": self._get_domain_device_instances,
            "device_config": self._get_domain_device_configs,
        }
        tasks = [reader_map[item_type](domain) for item_type in item_types]
        results = await gather(*tasks)

        # Flatten results into a single list
        result = [item for group in results for item in group]
        return result

    async def list_named_items(
            self, domain: str, item_type: str, simple_name: str
    ) -> list[dict[str, any]]:
        """
        List items in domain which match item_type and simple_name

        :param domain: domain to list items from
        :param item_type: item_type to match
        :param simple_name: name to match
        :return: a list of dicts where each entry has keys: uuid, item_type
                 and simple_name
        """
        items = await self.list_items(domain, [item_type])
        return [item for item in items
                if item.get('simple_name') == simple_name]

    async def list_domains(self):
        return [d.name for d in await self._execute_all(select(ProjectDomain))]

    async def get_devices_from_domain(
            self, domain: str) -> list[dict[str, any]]:
        """Return an information about the device topology """
        query = (
            select(DeviceInstance, DeviceServer, Project)
            .join(DeviceInstance.device_server)
            .join(DeviceServer.project)
            .join(Project.project_domain)
            .where(ProjectDomain.name == domain)
            .where(Project.is_trashed.is_(False))
        )
        async with self.session_gen() as session:
            result = await session.exec(query)
            rows = result.all()
            instances = []
            for device_instance, device_server, project in rows:
                instances.append({
                    "device_uuid": device_instance.uuid,
                    "device_name": device_instance.name,
                    "device_class": device_instance.class_id,
                    "server_name": device_server.name,
                    "project_uuid": project.uuid,
                    "project_name": project.name,
                })

            return instances

    async def _emit_project_item(self, uuid: str) -> dict[str, any] | None:
        """Internal emitter method to load a project item"""
        project = await self._get_project_from_uuid(uuid)
        if not project:
            raise ProjectDBError(f"Project item with UUID {uuid} not found.")
        subprojects, scenes, macros, servers = await gather(
            self._get_subprojects_of_project(project),
            self._get_scenes_of_project(project),
            self._get_macros_of_project(project),
            self._get_servers_of_project(project)
        )
        await self._register_project_load(project)
        item = {
            "uuid": uuid,
            "xml": emit_project_xml(
                project, scenes, macros, servers, subprojects)}
        return item

    async def _emit_scene_item(self, uuid: str) -> dict[str, any] | None:
        scene = await self._get_scene_from_uuid(uuid)
        if not scene:
            raise ProjectDBError(f"Scene item with UUID {uuid} not found.")
        item = {"uuid": uuid,
                "xml": emit_scene_xml(scene)}
        return item

    async def _emit_macro_item(self, uuid: str) -> dict[str, any] | None:
        macro = await self._get_macro_from_uuid(uuid)
        if not macro:
            raise ProjectDBError(f"Macr item with UUID {uuid} not found.")
        item = {"uuid": uuid,
                "xml": emit_macro_xml(macro)}
        return item

    async def _emit_server_item(self, uuid: str) -> dict[str, any] | None:
        server = await self._get_server_from_uuid(uuid)
        if not server:
            raise ProjectDBError(f"Server item with UUID {uuid} not found.")
        instances = await self._get_devices_of_server(server)
        item = {"uuid": uuid,
                "xml": emit_server_xml(server, instances)}
        return item

    async def _emit_device_item(self, uuid: str) -> dict[str, any] | None:
        instance = await self._get_device_from_uuid(uuid)
        if not instance:
            raise ProjectDBError(f"Device item with UUID {uuid} not found.")
        configs = await self._get_configs_of_device(instance)
        item = {"uuid": uuid,
                "xml": emit_device_xml(instance, configs)}
        return item

    async def _emit_config_item(self, uuid: str) -> dict[str, any] | None:
        config = await self._get_config_from_uuid(uuid)
        if not config:
            raise ProjectDBError(f"Config item with UUID {uuid} not found.")
        item = {"uuid": uuid,
                "xml": emit_config_xml(config)}
        return item

    async def load_item(
            self, domain: str, items: list[dict]) -> list[dict[str, any]]:
        """
        Loads a list of items of a specified type from a given domain.

        :param domain: domain name (not used for SQL; legacy from ExistDB)
        :param items_uuids: list of UUIDs to load
        :param item_type: optional type of the items to be loaded
        :return: List of dicts with keys "uuid" and "xml"
        :raises: ProjectDBError if any item can't be found
        """
        reader_map: dict[str, Callable[[str], Awaitable[dict[str, any]]]] = {
            "project": self._emit_project_item,
            "macro": self._emit_macro_item,
            "scene": self._emit_scene_item,
            "device_server": self._emit_server_item,
            "device_instance": self._emit_device_item,
            "device_config": self._emit_config_item,
        }
        futures = [reader_map[item["item_type"]](item["uuid"])
                   for item in items]
        loaded_items = await gather(*futures)
        return loaded_items

    async def _check_for_modification(self, uuid: str, client_date: str,
                                      item_type: str) -> tuple[bool, str]:
        """ Check whether the item with of the given `domain` and `uuid` was
        modified

        :param uuid: the item's uuid
        :param client_date: the item's last modified time stamp
        :param item_type: the type of the item to be checked (e.g. 'project')
        :return: A tuple stating whether the item was modified inbetween and a
                 string describing the reason
        """
        last_modified = None
        match item_type:
            case "project":
                project = await self._get_project_from_uuid(uuid)
                if project:
                    last_modified = project.date
            case "macro":
                macro = await self._get_macro_from_uuid(uuid)
                if macro:
                    last_modified = macro.date
            case "scene":
                scene = await self._get_scene_from_uuid(uuid)
                if scene:
                    last_modified = scene.date
            case "device_server":
                server = await self._get_server_from_uuid(uuid)
                if server:
                    last_modified = server.date
            case "device_instance":
                instance = await self._get_device_from_uuid(
                    uuid)
                if instance:
                    last_modified = instance.date
            case "device_config":
                config = await self._get_config_from_uuid(uuid)
                if config:
                    last_modified = config.date

        if last_modified:
            client_modified = datetime_from_str(client_date)
            # `last_modified`` was read from the DB as a naive datetime, but we
            # know it is in UTC. Add the timezone information so it can be
            # compared to the non-naive obtained from client_date.
            last_modified = last_modified.replace(tzinfo=UTC)
            if client_modified < last_modified:
                # The data has been changed in the database between the
                # client_date that the GUI client sent and the current time.
                # Need to convert UTC timestamps to local timestamps
                lm = date_utc_to_local(last_modified)
                cm = date_utc_to_local(client_modified)
                text = (
                    f'Item of type "{item_type}" with uuid "{uuid}" '
                    f'has been modified in the DB ({lm}) '
                    f'after its last retrieval by the client ({cm}).')
                return True, text

        return False, ""

    async def save_item(self, domain: str, uuid: str, item_xml: str,
                        overwrite: bool = False) -> str:
        """
        Saves a item xml file into the database. It will
        create a new entry if the item does not exist yet, or update the item
        if it does exist.

        The root node of the xml should contain a `item_type` entry identifying
        the type of the item as one of the following:

        'projects', 'scenes', 'macros', 'device_configs', 'device_servers'

        If domain does not exist in the database a RuntimeError is raised.

        :param domain: the domain under which this item is to be stored
        :param uuid: the item's uuid
        :param item_xml: the xml containing the item information
        :param overwrite (unused): defaults to False. If set to True versioning
                                   information is removed prior to database
                                   injection, allowing to overwrite in case of
                                   versioning conflicts.

        :return: date (str) of saving


        :raises: ProjectDBError on failure.
        """
        del overwrite  # To avoid unintentional use of unused parameter

        # Comply with the ExistDB back-end, which creates the domain for the
        # project being saved if the domain doesn't exist yet. The add_domain
        # method has the same "ensure domain exists" semantics of its overriden
        # counterpart in DatabaseBase.
        await self.add_domain(domain)

        # Extract some information
        try:
            # NOTE: The client might send us garbage
            item_tree = make_xml_if_needed(item_xml)
        except ValueError:
            msg = f'XML parse error for item "{uuid}"'
            raise ProjectDBError(msg)

        # XXX: Cannot be `None`
        item_type = item_tree.attrib['item_type']
        if not item_tree.attrib.get('date'):
            timestamp = datetime_str_now()
            item_tree.attrib['date'] = timestamp
        else:
            modified, reason = await self._check_for_modification(
                uuid, item_tree.attrib['date'], item_type)
            if modified:
                simple_name = item_tree.attrib.get('simple_name', '')
                message = (
                    f"The <b>{item_type}</b> item <b>{simple_name}</b> "
                    f"could not be saved: {reason}")
                raise ProjectDBError(message)
            # Update time stamp
            timestamp = datetime_str_now()
            item_tree.attrib['date'] = timestamp

        # XXX: Add a revision/alias to keep old code from blowing up
        # Is this needed?
        item_tree.attrib['revision'] = '0'
        item_tree.attrib['alias'] = 'default'

        # And back conversion to string
        item_xml = make_str_if_needed(item_tree)

        match item_type:
            case "project":
                await self._save_project_item(
                    domain, uuid, item_xml, timestamp)
            case "macro":
                await self._save_macro_item(
                    uuid, item_xml, timestamp)
            case "scene":
                await self._save_scene_item(
                    uuid, item_xml, timestamp)
            case "device_server":
                await self._save_device_server_item(
                    uuid, item_xml, timestamp)
            case "device_instance":
                await self._save_device_instance_item(
                    uuid, item_xml, timestamp)
            case "device_config":
                await self._save_device_config_item(
                    uuid, item_xml, timestamp)
            case _:
                raise ProjectDBError(
                    f"Saving of items of type '{item_type}' not supported")

        return timestamp

    async def _get_configuration_from_device_part(
            self, domain: str, name_part: str,
            only_active: bool = False
    ) -> list[dict[str, str]]:
        """
        Returns a list of configurations for a given device

        :param domain: DB domain
        :param name_part: part of device name; search is case-insensitive.
        :param only_active: if True only returns the active configurations

        :return: a list of dicts:

            [{"config_id": uuid of the configuration,
              "device_uuid": device instance uuid in the DB,
              "device_id": device instance id},
              ...
            ]
        """
        instances = await self._get_domain_devices_by_name_part(
            domain, name_part
        )
        results = []
        for instance in instances:
            configs = await self._get_configs_of_device(instance)
            filtered = [{"config_id": config.uuid,
                         "device_uuid": instance.uuid,
                         "device_id": instance.name}
                        for config in configs
                        if config.is_active or not only_active]
            results.extend(filtered)

        return results

    async def _get_configurations_from_device_name(
            self, domain: str, instance_id: str
    ) -> list[dict[str, str]]:
        """
        Returns a list of active configurations for a given device_id

        :param domain: DB domain
        :param instance_id: instance id of the device
        :return: a list of dicts:
            [{"configid": uuid of the configuration,
              "instanceid": device instance uuid in the DB},
              ...
            ]
        """
        configs = await self._get_configuration_from_device_part(
            domain, instance_id, only_active=True)

        items = [{"configid": config["config_id"],
                  "instanceid": config["device_uuid"]}
                 for config in configs]
        return items

    async def _get_project_from_device_uuid(
            self, domain: str, uuid: str) -> Project:
        """
        Returns the project which contain a device instance with a given uuid

        :param domain: DB domain
        :param uuid: the uuid of the device instance from the database
        :return: a list containing project names, uuids and last modification
                 date.
        """
        instance = await self._get_device_from_uuid(uuid)
        # From the DB structure, a given device instance only belongs to
        # one project
        projects = await self._execute_first(
            select(Project).join(DeviceServer).where(
                DeviceServer.id == instance.device_server_id))
        return projects

    async def update_trashed(self, **info) -> None:
        """ Update trashed attribute for project

        :raises: ProjectDBError on failure
        """
        item_uuid = info['uuid']
        item_type = info['item_type']
        assert item_type == "project", "Only projects can be trashed"
        value = info["value"]
        assert isinstance(value, bool)

        model = await self._get_project_from_uuid(
            item_uuid)
        if model is None:
            raise ProjectDBError(
                f"No item of type '{item_type}' with UUID '{item_uuid}' "
                "found in the database")

        model.is_trashed = value
        async with self.session_gen() as session:
            session.add(model)
            await session.commit()

    async def get_device_config_from_device_uuid(self, device_uuid: str):
        device_instance = await self._execute_first(
            select(DeviceInstance)
            .options(selectinload(DeviceInstance.configs))
            .where(DeviceInstance.uuid == device_uuid))
        if not device_instance:
            return None

        # Find the first config with name "default"
        configuration = next(
            (config.config_data for config in device_instance.configs
             if config.name == "default"), None)

        return configuration

    async def get_projects_with_conf(self, domain, device_id):
        """
        Returns a dict with projects and active configurations from a device
        name.

        :param domain: DB domain
        :param device_id: the device to return the information for.
        :return: a dict:
            {"project name": configuration uuid,
             ...}
        """
        configs = await self._get_configurations_from_device_name(
            domain, device_id)
        projects = dict()
        for config in configs:
            instance_id = config["instanceid"]
            project = await self._get_project_from_device_uuid(
                domain, instance_id)
            projects[project.name] = config["configid"]
        return projects

    async def get_projects_with_device(
            self, domain: str, name_part: str) -> list[dict[str, any]]:
        """
        Returns a list of dictionaries with data about projects from a given
        domain that contain active configurations for a given device.

        :param domain: domain of the projects to be returned
        :param name_part: part of name of devices for which project data
                               must be returned.
        :return: a list of dicts:
            [{"project_name": name of project,
              "date": last modification timestamp for the project,
              "uuid": uuid of project,
              "devices": list of ids of prj devices with the given part},
            ...]
        """
        items = await self._execute_all(
            select(Project, DeviceInstance)
            .join(ProjectDomain).join(DeviceServer).join(DeviceInstance)
            .where(ProjectDomain.name == domain)
            .where(Project.is_trashed.is_(False))
            .filter(DeviceInstance.name.ilike(f'%{name_part}%'))
        )

        projects = {}
        for item in items:
            if item.Project.id not in projects.keys():
                projects[item.Project.id] = {
                    "project_name": item.Project.name,
                    "date": datetime_to_str(item.Project.date),
                    "uuid": item.Project.uuid,
                    "items": []}
            projects[item.Project.id]["items"].append(
                item.DeviceInstance.name)
        return list(projects.values())

    async def get_projects_with_server(
            self, domain: str, name_part: str) -> list[dict[str, any]]:
        """
        Returns a list of dictionaries with data about projects that
        contain macros.

        :param domain: domain of the projects to be returned.
        :param name_part: part of name of macro for which project data
                          must be returned.
        :return: a list of dicts:
            [
            {"project_name": name of project,
             "date": last modification timestamp for the project,
             "uuid": uuid of project
             "items": list of ids of prj servers with the given part},
             ...]
        """
        items = await self._execute_all(
            select(Project, DeviceServer)
            .join(ProjectDomain).join(DeviceServer)
            .where(ProjectDomain.name == domain)
            .where(Project.is_trashed.is_(False))
            .filter(DeviceServer.name.ilike(f'%{name_part}%')))

        projects = {}
        for item in items:
            if item.Project.id not in projects.keys():
                projects[item.Project.id] = {
                    "project_name": item.Project.name,
                    "date": datetime_to_str(item.Project.date),
                    "uuid": item.Project.uuid,
                    "items": []}
            projects[item.Project.id]["items"].append(
                item.DeviceServer.name)
        return list(projects.values())

    async def get_projects_with_macro(
            self, domain: str, name_part: str) -> list[dict[str, any]]:
        """
        Returns a list of dictionaries with data about projects that
        contain macros.

        :param domain: domain of the projects to be returned
        :param name_part: part of name of macro for which project data
                          must be returned.
        :return: a list of dicts:
            [
            {"project_name": name of project,
             "date": last modification timestamp for the project,
             "uuid": uuid of project
             "macros": list of ids of prj devices with the given part},
             ...]
        """
        items = await self._execute_all(
            select(Project, Macro)
            .join(ProjectDomain).join(Macro)
            .where(ProjectDomain.name == domain)
            .where(Project.is_trashed.is_(False))
            .filter(Macro.name.ilike(f'%{name_part}%')))

        projects = {}
        for item in items:
            if item.Project.id not in projects.keys():
                projects[item.Project.id] = {
                    "project_name": item.Project.name,
                    "date": datetime_to_str(item.Project.date),
                    "uuid": item.Project.uuid,
                    "items": []}
            projects[item.Project.id]["items"].append(
                item.Macro.name)
        return list(projects.values())

    # region Reader

    async def _execute_all(self, base_query, order_by=None):
        if order_by is not None:
            base_query = base_query.order_by(order_by)
        async with self.session_gen() as session:
            result = await session.exec(base_query)
            return result.all()

    async def _execute_first(self, base_query, order_by=None):
        if order_by is not None:
            base_query = base_query.order_by(order_by)
        async with self.session_gen() as session:
            result = await session.exec(base_query)
            return result.first()

    # endregion
    # region Domain

    async def _get_domain_projects(self, domain: str):
        result = await self._execute_all(
            select(Project).join(ProjectDomain).where(
                ProjectDomain.name == domain))
        return [{
            "uuid": p.uuid,
            "item_type": "project",
            "simple_name": p.name,
            "is_trashed": p.is_trashed,
            "date": datetime_to_str(p.date),
        } for p in result]

    async def _get_domain_macros(self, domain):
        items = await self._execute_all(
            select(Macro)
            .join(Project).join(ProjectDomain)
            .where(ProjectDomain.name == domain))
        return [{"uuid": i.uuid, "item_type": "macro",
                 "simple_name": i.name, "date": datetime_to_str(i.date)}
                for i in items]

    async def _get_domain_scenes(self, domain: str):
        items = await self._execute_all(
            select(Scene)
            .join(Project).join(ProjectDomain)
            .where(ProjectDomain.name == domain))
        return [{"uuid": i.uuid, "item_type": "scene",
                 "simple_name": i.name, "date": datetime_to_str(i.date)}
                for i in items]

    async def _get_domain_device_servers(self, domain: str):
        items = await self._execute_all(
            select(DeviceServer)
            .join(Project).join(ProjectDomain)
            .where(ProjectDomain.name == domain)
        )
        return [{"uuid": i.uuid, "item_type": "device_server",
                 "simple_name": i.name, "date": datetime_to_str(i.date)}
                for i in items]

    async def _get_domain_device_instances(self, domain):
        items = await self._execute_all(
            select(DeviceInstance)
            .join(DeviceServer).join(Project).join(ProjectDomain)
            .where(ProjectDomain.name == domain))
        return [{"uuid": i.uuid, "item_type": "device_instance",
                 "simple_name": i.name, "date": datetime_to_str(i.date)}
                for i in items]

    async def _get_domain_device_configs(self, domain: str):
        items = await self._execute_all(
            select(DeviceConfig)
            .join(DeviceInstance).join(DeviceServer).join(Project)
            .join(ProjectDomain)
            .where(ProjectDomain.name == domain))
        return [{"uuid": i.uuid, "item_type": "device_config",
                 "simple_name": i.name, "date": datetime_to_str(i.date)}
                for i in items]

    # endregion

    async def _get_subprojects_of_project(self, project: Project):
        links = await self._execute_all(
            select(ProjectSubproject).where(
                ProjectSubproject.project_id == project.id),
            order_by=ProjectSubproject.order)
        subprojects = []
        futures = [self._execute_first(select(Project).where(
            Project.id == link.subproject_id)) for link in links]
        if futures:
            subprojects = await gather(*futures)
            subprojects = [sub for sub in subprojects if sub]
        return subprojects

    async def _get_scenes_of_project(self, project):
        return await self._execute_all(
            select(Scene).where(Scene.project_id == project.id),
            order_by=Scene.order)

    async def _get_macros_of_project(self, project):
        return await self._execute_all(
            select(Macro).where(Macro.project_id == project.id),
            order_by=Macro.order)

    async def _get_servers_of_project(self, project):
        return await self._execute_all(
            select(DeviceServer).where(DeviceServer.project_id == project.id),
            order_by=DeviceServer.order)

    async def _get_devices_of_server(self, server):
        return await self._execute_all(
            select(DeviceInstance).where(
                DeviceInstance.device_server_id == server.id),
            order_by=DeviceInstance.order)

    async def _get_configs_of_device(self, instance):
        return await self._execute_all(
            select(DeviceConfig).where(
                DeviceConfig.device_instance_id == instance.id),
            order_by=DeviceConfig.order)

    # region UUID

    async def _get_project_from_uuid(self, uuid):
        project = await self._execute_first(
            select(Project).where(Project.uuid == uuid))
        return project

    async def _get_scene_from_uuid(self, uuid):
        return await self._execute_first(
            select(Scene).where(Scene.uuid == uuid))

    async def _get_macro_from_uuid(self, uuid):
        return await self._execute_first(
            select(Macro).where(Macro.uuid == uuid))

    async def _get_server_from_uuid(self, uuid):
        return await self._execute_first(
            select(DeviceServer).where(DeviceServer.uuid == uuid))

    async def _get_device_from_uuid(self, uuid):
        return await self._execute_first(
            select(DeviceInstance).where(DeviceInstance.uuid == uuid))

    async def _get_config_from_uuid(self, uuid):
        return await self._execute_first(
            select(DeviceConfig).where(DeviceConfig.uuid == uuid))

    # endregion
    # region namepart

    async def _get_domain_devices_by_name_part(
            self, domain: str, name_part: str):
        return await self._execute_all(
            select(DeviceInstance)
            .join(DeviceServer).join(Project).join(ProjectDomain)
            .where(ProjectDomain.name == domain)
            .filter(DeviceInstance.name.ilike(f'%{name_part}%')))

    # end region
    # region Writer

    async def _register_project_load(self, project: Project):
        async with self.session_gen() as session:
            project.last_loaded = datetime_now()
            session.add(project)
            await session.commit()

    async def _save_project_item(
            self, domain: str, uuid: str, xml: str, timestamp: str):
        project_xml = etree.fromstring(xml)
        date = datetime_from_str(timestamp)

        # Save the project attributes
        project_id = None  # the id of the updated or new project
        async with self.session_gen() as session:
            result = await session.exec(
                select(ProjectDomain).where(ProjectDomain.name == domain))
            project_domain = result.first()
            # domain is expected to correspond to a valid domain.
            assert project_domain is not None

            result = await session.exec(
                select(Project).where(Project.uuid == uuid))
            project = result.first()

            attrs = project_xml.attrib
            if not project:
                # There's still no record for the project in the DB
                trashed = get_trashed(attrs.get('is_trashed'))
                project = Project(
                    uuid=uuid,
                    name=attrs['simple_name'],
                    is_trashed=trashed,
                    date=date,
                    project_domain_id=project_domain.id)
                session.add(project)
                # The flush sequence is needed for the case of a
                # new project. A new project only gets its ID after being
                # stored in the database. This is a save marker
                await session.flush()
            else:
                # There's a record for the project in the DB - update its
                # attributes from the data in the xml
                project.name = attrs['simple_name']
                project.is_trashed = get_trashed(attrs['is_trashed'])
                project.date = date
                project.project_domain_id = project_domain.id
                session.add(project)

            # Save the children references to the project and
            # their relative order. For each child type, the items of that
            # type currently linked to the project must be first unlinked
            # to the project.
            project_elements = (
                project_xml.getchildren()[0].getchildren()[0]
                if project_xml.getchildren()
                and project_xml.getchildren()[0].getchildren() else None)

            group_map = {}
            if project_elements is not None:
                group_map = {child.tag.lower(): child
                             for child in project_elements.getchildren()}

            # several child elements
            macro_group = group_map.get("macros", None)
            scene_group = group_map.get("scenes", None)
            servers_group = group_map.get("servers", None)
            subprojects_group = group_map.get("subprojects", None)

            project_id = project.id

            # --- Macros ---
            result = await session.exec(
                select(Macro).where(Macro.project_id == project_id))
            existing = result.all()
            for macro in existing:
                macro.project_id = None
                macro.order = 0
                session.add(macro)

            macro_index = 0
            if macro_group is not None:
                for macro_elem in macro_group.getchildren():
                    macro_uuid = macro_elem.getchildren()[0].text
                    result = await session.exec(
                        select(Macro).where(Macro.uuid == macro_uuid))
                    macro = result.first()
                    if not macro:
                        raise ProjectDBError(
                            f'Macro with uuid "{macro_uuid}" not found in the '
                            'database. Cannot link the macro to project '
                            f'"{project.name}" ({project.uuid})')
                    macro.project_id = project_id
                    macro.order = macro_index
                    session.add(macro)
                    macro_index += 1

            # --- Scenes ---
            result = await session.exec(
                select(Scene).where(Scene.project_id == project_id))
            existing = result.all()
            for scene in existing:
                scene.project_id = None
                scene.order = 0
                session.add(scene)

            if scene_group is not None:
                scene_index = 0
                for scene_element in scene_group.getchildren():
                    scene_uuid = (
                        scene_element.getchildren()[0].text
                        if len(scene_element.getchildren()) > 0 else None)
                    # The legacy unit tests had the scene UUID as an attribute
                    # of a <xml> tag
                    if (scene_uuid is None
                            and "uuid" in scene_element.attrib.keys()):
                        scene_uuid = scene_element.attrib['uuid']

                    result = await session.exec(
                        select(Scene).where(Scene.uuid == scene_uuid))
                    scene = result.first()
                    if not scene:
                        raise ProjectDBError(
                            f'Scene with uuid "{scene_uuid}" not found in the '
                            'database. Cannot link the scene to project '
                            f'"{project.name}" ({project.uuid})')
                    scene.project_id = project_id
                    scene.order = scene_index
                    session.add(scene)
                    scene_index += 1

            # --- Servers ---

            result = await session.exec(
                select(DeviceServer).where(
                    DeviceServer.project_id == project_id))
            existing = result.all()
            # erase order and reference
            for server in existing:
                server.project_id = None
                server.order = 0
                session.add(server)

            if servers_group is not None:
                server_index = 0
                for server_element in servers_group.getchildren():
                    server_uuid = server_element.getchildren()[0].text
                    result = await session.exec(
                        select(DeviceServer).where(
                            DeviceServer.uuid == server_uuid))
                    server = result.first()
                    if not server:
                        raise ProjectDBError(
                            f'Server with uuid "{server_uuid}" not found in '
                            'the database. Cannot link the server to project'
                            f'"{project.name}" ({project.uuid})')
                    server.project_id = project_id
                    server.order = server_index
                    session.add(server)
                    server_index += 1

            # --- Subprojects ---

            # Initially all the previously saved subproject are to be
            # potentially removed
            # Note: subprojects are only a reference to projects, not real
            #       entities stored in the database like scenes and macros.
            result = await session.exec(
                select(ProjectSubproject).where(
                    ProjectSubproject.project_id == project_id))
            existing = result.all()
            subprojects_to_delete = {
                subproject.subproject_id for subproject in existing}

            # Iterate over the subprojects to be saved
            if subprojects_group is not None:
                subproject_index = 0
                for sub_element in subprojects_group.getchildren():
                    subproject_uuid = sub_element.getchildren()[0].text
                    result = await session.exec(
                        select(Project).where(
                            Project.uuid == subproject_uuid))
                    subproject = result.first()
                    if not subproject:
                        raise ProjectDBError(
                            f'Subproject with uuid "{server_uuid}" not found '
                            'in the database. Cannot link the subproject to '
                            f'project "{project.name}" ({project.uuid})')
                    # Search for the subproject among the current projects, and
                    # either add it, update it or leave it in the list of
                    # subproject to be removed
                    if subproject.id in subprojects_to_delete:
                        # The project was in the DB and should be kept - just
                        # update it
                        subprojects_to_delete.remove(subproject.id)
                        query = select(ProjectSubproject).where(
                            ProjectSubproject.project_id == project_id,
                            ProjectSubproject.subproject_id == subproject.id)
                        result = await session.exec(query)
                        subproject_update = result.first()
                        if subproject_update:
                            subproject_update.order = subproject_index
                        session.add(subproject_update)
                    else:
                        # The project wasn't in the DB - add it
                        subproject_new = ProjectSubproject(
                            project_id=project_id,
                            subproject_id=subproject.id,
                            order=subproject_index)
                        session.add(subproject_new)
                    subproject_index += 1

            # Remove the subprojects that were in the DB before but shouldn't
            # be there after the save
            for subproject_id in subprojects_to_delete:
                result = await session.exec(
                    select(ProjectSubproject).where(
                        ProjectSubproject.project_id == project_id,
                        ProjectSubproject.subproject_id == subproject_id))
                to_delete = result.first()
                if to_delete:
                    await session.delete(to_delete)

            await session.commit()

    async def _save_macro_item(self, uuid: str, xml: str, timestamp: str):
        macro_obj = etree.fromstring(xml)
        macro_name = macro_obj.attrib["simple_name"]
        # In MySQL the macro bodies are not Base64 encoded
        macro_body = base64.b64decode(macro_obj.getchildren()[0].text)

        date = datetime_from_str(timestamp)

        async with self.session_gen() as session:
            result = await session.exec(
                select(Macro).where(Macro.uuid == uuid))
            macro = result.first()
            if macro:
                # A macro is being updated
                macro.name = macro_name
                macro.date = date
                macro.body = macro_body
            else:
                # The macro is new
                macro = Macro(
                    uuid=uuid,
                    name=macro_name,
                    date=date,
                    body=macro_body)
            session.add(macro)
            await session.commit()

    async def _save_scene_item(self, uuid: str, xml: str, timestamp: str):
        date = datetime_from_str(timestamp)
        scene_obj = etree.fromstring(xml)
        scene_name = scene_obj.attrib["simple_name"]
        scene_svg_data = (
            etree.tostring(scene_obj.getchildren()[0]).decode("UTF-8")
            if len(scene_obj.getchildren()) > 0 else "")

        async with self.session_gen() as session:
            result = await session.exec(
                select(Scene).where(Scene.uuid == uuid))
            scene = result.first()
            if scene:
                # A scene is being updated

                # Remove all the scene links originated from the scene
                query_links = (
                    select(SceneLinkedScene)
                    .where(SceneLinkedScene.scene_id == scene.id))
                result_links = await session.exec(query_links)
                links = result_links.all()
                for link in links:
                    await session.delete(link)

                scene.name = scene_name
                scene.date = date
                scene.svg_data = scene_svg_data
            else:
                # The scene is new
                scene = Scene(
                    uuid=uuid,
                    name=scene_name,
                    date=date,
                    svg_data=scene_svg_data)
            session.add(scene)
            await session.commit()
            if scene.id is None:
                await session.refresh(scene)
            await self._save_scene_links(scene)

    async def _save_scene_links(self, scene: Scene):
        target_uuids = get_scene_links(scene)
        for uuid in target_uuids:
            target_scene = await self._get_scene_from_uuid(uuid)
            if target_scene is None:
                # XXX: Here we are graceful
                logger.error(
                    f"Target scene with UUID '{uuid}' doesn't exist in "
                    f"the database. Skipping scene link registration.")
                continue
            async with self.session_gen() as session:
                try:
                    linked_scene = SceneLinkedScene(
                        scene_id=scene.id,
                        linked_scene_id=target_scene.id)
                    session.add(linked_scene)
                    await session.commit()
                except IntegrityError:
                    # A scene can have multiple scene links to the same scene.
                    # An attempt to insert the second repeated link will
                    # violate the unique constraint for the combination of
                    # source and target scene of the link. This is not an error
                    # and can be ignored. The unique constraint should be kept
                    # though as it prevents storage of repeated information.
                    await session.rollback()

    async def _save_device_config_item(
            self, uuid: str, xml: str, timestamp: str):
        date = datetime_from_str(timestamp)
        config_obj = etree.fromstring(xml)
        config_name = config_obj.attrib["simple_name"]
        config_data = (
            etree.tostring(config_obj.getchildren()[0]).decode("UTF-8")
            if len(config_obj.getchildren()) > 0 else "")

        async with self.session_gen() as session:
            result = await session.exec(
                select(DeviceConfig).where(DeviceConfig.uuid == uuid))
            config = result.first()
            if config:
                # A device config is being updated
                config.name = config_name
                config.config_data = config_data
                config.date = date
            else:
                # A new device config
                config = DeviceConfig(
                    uuid=uuid,
                    name=config_name,
                    config_data=config_data,
                    date=date)

            session.add(config)
            await session.commit()

    async def _save_device_instance_item(
            self, uuid: str, xml: str, timestamp: str):
        # XXX: Why is the uuid not taken here?
        date = datetime_from_str(timestamp)
        instance_element = etree.fromstring(xml)

        instance_tag = instance_element.getchildren()[0]
        instance_id = instance_tag.attrib['instance_id']
        class_id = instance_tag.attrib['class_id']
        active_uuid = instance_tag.attrib['active_uuid']
        config_objs = instance_tag.getchildren()

        # Insert or update the device instance in the DB
        async with self.session_gen() as session:
            result = await session.exec(
                select(DeviceInstance).where(DeviceInstance.uuid == uuid))
            instance = result.first()
            if instance:
                # Updates the device instance
                instance.name = instance_id
                instance.class_id = class_id
                instance.date = date
                session.add(instance)
            else:
                # The device instance must be added to the DB
                instance = DeviceInstance(
                    uuid=uuid,
                    name=instance_id,
                    class_id=class_id,
                    date=date)
                session.add(instance)
                await session.flush()

            # Saves the device configs linked to the device instance that
            # has been just saved. First the currently linked configs
            # must be unlinked
            result = await session.exec(
                select(DeviceConfig).where(
                    DeviceConfig.device_instance_id == instance.id))
            existing = result.all()
            for config in existing:
                config.device_instance_id = None
                config.order = 0
                session.add(config)

            config_index = 0
            for config_obj in config_objs:
                config_obj_uuid = config_obj.attrib["uuid"]
                result = await session.exec(
                    select(DeviceConfig).where(
                        DeviceConfig.uuid == config_obj_uuid))
                config = result.first()
                if not config:
                    raise ProjectDBError(
                        f'Device config with uuid "{config_obj_uuid}" not '
                        "found in the database. Cannot link the config to "
                        f'instance "{instance.name}" ({instance.uuid})')
                config.device_instance_id = instance.id
                config.order = config_index
                config.is_active = (active_uuid == config_obj_uuid)
                config_index += 1
                session.add(config)

            await session.commit()

    async def _save_device_server_item(
            self, uuid: str, xml: str, timestamp: str):
        date = datetime_from_str(timestamp)
        server_obj = etree.fromstring(xml)
        server_name = server_obj.attrib["simple_name"]
        server_tag = (server_obj.getchildren()[0]
                      if len(server_obj.getchildren()) > 0 else None)
        instance_elements = (server_tag.getchildren() if server_tag is not None
                             else [])

        async with self.session_gen() as session:
            result = await session.exec(
                select(DeviceServer).where(DeviceServer.uuid == uuid))
            server = result.first()
            if server:
                # An existing device server is being saved
                server.name = server_name
                server.date = date
                session.add(server)
            else:
                # A new device server is being saved
                server = DeviceServer(
                    uuid=uuid,
                    name=server_name,
                    date=date)
                session.add(server)
                await session.flush()

            # Saves the device instances linked to the device server just saved
            # First unlinks all the currently linked device instances.
            result = await session.exec(
                select(DeviceInstance).where(
                    DeviceInstance.device_server_id == server.id))
            existing = result.all()
            for instance in existing:
                instance.device_server_id = None
                instance.order = 0
                session.add(instance)

            instance_index = 0
            for instance_element in instance_elements:
                instance_element_uuid = instance_element.attrib["uuid"]
                result = await session.exec(
                    select(DeviceInstance).where(
                        DeviceInstance.uuid == instance_element_uuid))
                instance = result.first()
                if not instance:
                    raise ProjectDBError(
                        f'Device with uuid "{instance_element_uuid}" not '
                        "found in the database. Cannot link the instance to "
                        f'server "{server.name}" ({server.uuid})')
                instance.device_server_id = server.id
                instance.order = instance_index
                instance_index += 1
                session.add(instance)

            await session.commit()

    async def add_domain(self, domain: str):
        if await self.domain_exists(domain):
            return
        """Add a domain to the project database"""
        async with self.session_gen() as session:
            project_domain = ProjectDomain(name=domain)
            session.add(project_domain)
            await session.commit()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
import socket
from asyncio import (
    Future, IncompleteReadError, StreamReader, StreamWriter, WriteTransport,
    get_running_loop, sleep)
from struct import pack
from zlib import adler32

import pytest
import pytest_asyncio
from psutil import net_if_addrs

from karabo.middlelayer import (
    Channel, Hash, Int32, NetworkInput, NetworkOutput, Proxy, RingQueue,
    background, encodeBinary)
from karabo.middlelayer.pipeline import IP_PATTERN, get_hostname_from_interface
from karabo.middlelayer.testing import run_test


class ChecksumTransport(WriteTransport):
    def __init__(self):
        self.closed = False
        self.checksum = adler32(b"")
        self._extra = {"sockname": ("Local", 8080),
                       "peername": ("Remote", 8080)}

    def write(self, data):
        self.checksum = adler32(data, self.checksum)

    def close(self):
        self.closed = True

    def is_closing(self):
        return self.closed


class FakeTimestamp:
    """as the order of a dictionary is not defined in Python 3.4, we cannot
    use normal timestamps, as their serialization depends on dictionary order.
    This fake timestamp always serializes the same way"""

    def toDict(self):
        return {"a": 3}


class ChannelContext:

    sample_data = [(Hash("a", 5), FakeTimestamp()),
                   (Hash("b", 7), FakeTimestamp())]
    eos_channel = ""

    def __init__(self, channel) -> None:
        self.channel = channel
        self.reader = channel.reader
        self.writer = channel.writer

    def assertChecksum(self, checksum):
        assert self.writer.transport.checksum == checksum

    def feedHash(self, hsh):
        encoded = encodeBinary(hsh)
        self.reader.feed_data(pack("<I", len(encoded)))
        self.reader.feed_data(encoded)

    def feedRequest(self):
        self.feedHash(Hash("reason", "update"))

    async def write_something(self, output, number=10):
        output.channelName = "channelname"
        task = background(output.serve(self.reader, self.writer))
        for i in range(number):
            output.writeChunkNoWait(self.sample_data)
            await self.sleep()
        return task

    def feedTestData(self):
        encoded = encodeBinary(Hash("a", 7))
        meta = Hash("source", "a", "timestamp", False)
        self.feedHash(Hash("byteSizes", [len(encoded)] * 3,
                           "sourceInfo", [meta] * 3))
        self.reader.feed_data(pack("<I", 3 * len(encoded)))
        for i in range(3):
            self.reader.feed_data(encoded)

    def test_writeHash(self):
        h = Hash("b", 7)
        self.channel.writeHash(h)
        self.assertChecksum(124256394)

    async def handler(self, data, meta):
        self.data.append(data)
        self.meta.append(meta)

    async def eos_handler(self, name):
        self.eos_channel = name

    async def sleep(self):
        for _ in range(10):
            await sleep(0)


def test_ip_pattern():
    valid = ["192.168.1.1", "10.0.0.1", "172.16.0.1"]
    for ip_address in valid:
        assert IP_PATTERN.match(ip_address) is not None

    cidr_valid = ["192.168.1.1/24", "10.0.0.1/16", "172.16.0.1/12"]
    for ip_address in cidr_valid:
        assert IP_PATTERN.match(ip_address) is not None

    invalid = ["192.168.1", "192.168.1/2",
               "abc.def.ghi.jkl"]
    for ip_address in invalid:
        assert IP_PATTERN.match(ip_address) is None

    wrong_ips = ["192.168.1.1/33", "10.0.0.1/40", "172.16.0.1/-1",
                 "256.256.256.256", "192.168.1.1/asdf"]
    # These wrong ip's will raise on network resolve
    for ip_address in wrong_ips:
        assert IP_PATTERN.match(ip_address) is not None
        with pytest.raises(ValueError):
            get_hostname_from_interface(ip_address)


@pytest_asyncio.fixture(scope="function", loop_scope="module")
async def channelContext():
    loop = get_running_loop()
    reader = StreamReader(loop=loop)
    writer = StreamWriter(ChecksumTransport(), None, None, loop=loop)
    channel = Channel(reader, writer, "channelname")
    ctx = ChannelContext(channel)
    yield ctx


@run_test
def test_writeSize(channelContext):
    channelContext.channel.writeSize(218)
    channelContext.assertChecksum(57409755)


@run_test
async def test_ring_queue():
    size = 5
    ring = RingQueue(size)
    # Drop 0
    for i in range(size + 1):
        ring.put_nowait(i)

    assert ring.qsize() == size
    items = [ring.get_nowait() for _ in range(size)]
    assert items == [1, 2, 3, 4, 5]

    ring = RingQueue(size)
    for i in range(size + 1):
        ring.put_nowait(i)

    items = []
    for _ in range(size):
        item = await ring.get()
        items.append(item)
    assert items == [1, 2, 3, 4, 5]


@run_test
async def test_readBytes(channelContext):
    channelContext.reader.feed_data(b"\04\00\00\00abcd")
    data = await channelContext.channel.readBytes()
    assert data == b"abcd"


@run_test
async def test_readHash(channelContext):
    channelContext.feedHash(Hash("a", 5))
    data = await channelContext.channel.readHash()
    assert data["a"] == 5


@run_test
async def test_nextChunk(channelContext):
    future = Future()
    task = background(channelContext.channel.nextChunk(future))
    await channelContext.sleep()
    channelContext.assertChecksum(1)
    future.set_result(channelContext.sample_data)
    await channelContext.sleep()
    channelContext.assertChecksum(3020956469)
    assert not task.done()
    channelContext.feedHash(Hash("reason", "update"))
    await task


@run_test
async def test_nextChunk_eof(channelContext):
    future = Future()
    task = background(channelContext.channel.nextChunk(future))
    await channelContext.sleep()
    channelContext.reader.feed_eof()
    try:
        await task
    except IncompleteReadError as error:
        assert not error.partial
    else:
        channelContext.fail()


@run_test
async def test_processChunk(channelContext):
    network = NetworkInput({})
    network.handler = channelContext.handler
    network.raw = False
    await network._run()

    class Test(Proxy):
        a = Int32()

    task = background(network.processChunk(channelContext.channel, Test))
    channelContext.feedTestData()
    channelContext.data = []
    channelContext.meta = []
    assert (await task)
    assert len(channelContext.data) == 3
    assert channelContext.data[1].a == 7
    assert len(channelContext.meta) == 3
    assert channelContext.meta[2].source == "a"
    assert not channelContext.meta[0].timestamp


@run_test
async def test_processChunk_raw(channelContext):
    network = NetworkInput({})
    network.raw = True
    network.handler = channelContext.handler
    await network._run()
    task = background(network.processChunk(channelContext.channel, None))
    channelContext.feedTestData()
    channelContext.data = []
    channelContext.meta = []
    assert (await task)
    assert len(channelContext.data) == 3
    assert channelContext.data[1]["a"] == 7
    assert len(channelContext.meta) == 3
    assert channelContext.meta[2].source == "a"
    assert not channelContext.meta[0].timestamp


@run_test
async def test_processChunk_eos(channelContext):
    network = NetworkInput({})
    network.raw = True
    network.handler = channelContext.handler
    network.end_of_stream_handler = channelContext.eos_handler
    await network._run()
    assert channelContext.eos_channel == ""
    task = background(network.processChunk(channelContext.channel, None))
    channelContext.feedHash(Hash("endOfStream", True))
    channelContext.reader.feed_data(b"\0\0\0\0")
    channelContext.data = []
    channelContext.meta = []
    assert (await task)
    assert channelContext.data == []
    assert channelContext.eos_channel == "channelname"
    assert len(channelContext.meta) == 0


@run_test
async def test_processChunk_eof(channelContext, mocker):
    network = NetworkInput({})
    network.parent = mocker.Mock()
    await network._run()
    channelContext.reader.feed_eof()
    assert not (await network.processChunk(channelContext.channel, None))


@run_test
async def test_shared_drop(channelContext):
    output = NetworkOutput({"noInputShared": "drop"})
    await output._run()
    channelContext.feedHash(
        Hash("reason", "hello", "dataDistribution", "shared",
             "onSlowness", "drop", "instanceId", "Test"))
    task = await channelContext.write_something(output)
    await channelContext.sleep()
    channelContext.assertChecksum(3020956469)
    channelContext.feedRequest()
    await channelContext.sleep()
    channelContext.feedRequest()
    for _ in range(100):
        channelContext.assertChecksum(452019817)
        await channelContext.sleep()
    output.writeChunkNoWait(channelContext.sample_data)
    await channelContext.sleep()
    channelContext.assertChecksum(881158557)
    channelContext.reader.feed_eof()
    await task
    assert channelContext.writer.transport.closed


@run_test
async def test_shared_queue_drop(channelContext):
    output = NetworkOutput({"noInputShared": "queueDrop"})
    await output._run()
    channelContext.feedHash(
        Hash("reason", "hello", "dataDistribution", "shared",
             "onSlowness", "queueDrop", "instanceId", "Test"))
    task = await channelContext.write_something(output, 200)
    await channelContext.sleep()
    channelContext.assertChecksum(3020956469)
    channelContext.feedRequest()
    await channelContext.sleep()
    channelContext.feedRequest()
    await channelContext.sleep()
    for _ in range(100):
        channelContext.assertChecksum(881158557)
        await channelContext.sleep()
    output.writeChunkNoWait(channelContext.sample_data)
    await channelContext.sleep()
    channelContext.assertChecksum(881158557)
    for i in range(200):
        output.writeChunkNoWait(channelContext.sample_data)
    await channelContext.sleep()
    channelContext.assertChecksum(881158557)
    channelContext.reader.feed_eof()
    await task
    assert channelContext.writer.transport.closed


@run_test
def test_shared_throw(channelContext):
    with pytest.raises(ValueError):
        output = NetworkOutput({"noInputShared": "throw"})  # noqa


@run_test
async def test_shared_wait(channelContext):
    output = NetworkOutput({"noInputShared": "wait"})
    output.channelName = "channelname"
    await output._run()
    task = background(
        output.serve(
            channelContext.reader,
            channelContext.writer))
    channelContext.feedHash(
        Hash("reason", "hello", "dataDistribution", "shared",
             "onSlowness", "wait", "instanceId", "Test"))
    await channelContext.sleep()
    await output.writeChunk([(Hash("a", 5), FakeTimestamp())])
    await channelContext.sleep()
    channelContext.feedRequest()
    await channelContext.sleep()
    await output.writeChunk(channelContext.sample_data)
    await output.writeChunk(channelContext.sample_data)
    waiter = background(output.writeChunk(channelContext.sample_data))
    await channelContext.sleep()
    assert not waiter.done()
    channelContext.feedRequest()
    await channelContext.sleep()
    assert waiter.done()
    channelContext.assertChecksum(2194103602)
    channelContext.reader.feed_eof()
    await task
    assert channelContext.writer.transport.closed


@run_test
async def test_copy_drop(channelContext):
    output = NetworkOutput({"noInputShared": "drop"})
    await output._run()
    channelContext.feedHash(
        Hash("reason", "hello", "dataDistribution", "copy",
             "onSlowness", "drop", "instanceId", "Test"))
    task = await channelContext.write_something(output)
    await channelContext.sleep()
    channelContext.assertChecksum(3020956469)
    channelContext.feedRequest()
    await channelContext.sleep()
    for _ in range(100):
        channelContext.assertChecksum(3020956469)
        await channelContext.sleep()
    output.writeChunkNoWait(channelContext.sample_data)
    channelContext.feedRequest()
    await channelContext.sleep()
    channelContext.assertChecksum(452019817)
    channelContext.reader.feed_eof()
    await task
    assert channelContext.writer.transport.closed


@run_test
async def test_copy_queue_drop(channelContext):
    output = NetworkOutput({"noInputShared": "drop"})
    await output._run()
    channelContext.feedHash(
        Hash("reason", "hello", "dataDistribution", "copy",
             "onSlowness", "queueDrop", "instanceId", "Test"))
    task = await channelContext.write_something(output, 200)
    await channelContext.sleep()
    channelContext.assertChecksum(3020956469)
    # Write like crazy, we do not have a request and roll!
    output.writeChunkNoWait(channelContext.sample_data)
    output.writeChunkNoWait(channelContext.sample_data)
    output.writeChunkNoWait(channelContext.sample_data)
    channelContext.assertChecksum(3020956469)
    channelContext.feedRequest()
    await channelContext.sleep()
    for _ in range(100):
        channelContext.assertChecksum(452019817)
        await channelContext.sleep()
    output.writeChunkNoWait(channelContext.sample_data)
    channelContext.feedRequest()
    await channelContext.sleep()
    channelContext.assertChecksum(881158557)
    channelContext.reader.feed_eof()
    await task
    assert channelContext.writer.transport.closed


@run_test
async def test_copy_queue_drop_large_max_queue(channelContext):
    """Test the queue drop with a high max queue length"""
    output = NetworkOutput({"noInputShared": "queueDrop"})
    await output._run()
    channelContext.feedHash(
        Hash("reason", "hello", "dataDistribution", "copy",
             "onSlowness", "queueDrop", "instanceId", "Test",
             "maxQueueLength", 200))
    task = await channelContext.write_something(output, 200)
    await channelContext.sleep()
    channelContext.assertChecksum(3020956469)
    channelContext.reader.feed_eof()
    await task
    assert channelContext.writer.transport.closed


@run_test
async def test_copy_wait(channelContext):
    output = NetworkOutput({"noInputShared": "drop"})
    await output._run()
    output.channelName = "channelname"
    task = background(
        output.serve(
            channelContext.reader,
            channelContext.writer))
    channelContext.feedHash(
        Hash("reason", "hello", "dataDistribution", "copy",
             "onSlowness", "wait", "instanceId", "Test"))
    await output.writeChunk(channelContext.sample_data)
    await channelContext.sleep()
    await output.writeChunk(channelContext.sample_data)
    channelContext.feedRequest()
    await channelContext.sleep()
    await output.writeChunk(channelContext.sample_data)
    await output.writeChunk(channelContext.sample_data)
    waiter = background(output.writeChunk(channelContext.sample_data))
    await channelContext.sleep()
    assert not waiter.done()
    channelContext.feedRequest()
    await channelContext.sleep()
    assert waiter.done()
    channelContext.assertChecksum(881158557)
    channelContext.reader.feed_eof()
    await task
    assert channelContext.writer.transport.closed


@run_test
async def test_get_information(channelContext):
    net_segment = None
    ip_address = None
    for interface in net_if_addrs().values():
        for nic in interface:
            if nic.family != socket.AF_INET:
                continue
            ip_address = nic.address
            return ip_address

    assert net_segment is not None
    assert ip_address is not None
    output = NetworkOutput({"hostname": net_segment})
    await output._run()
    response = output.getInformation("channelName")
    assert "connectionType" in response
    assert "port" in response
    assert response["hostname"] == ip_address
    assert output.address == ip_address



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from contextlib import AbstractAsyncContextManager


class DatabaseBase(AbstractAsyncContextManager):

    async def initialize(self):
        """"Initialize the async interface for the database"""

    async def get_configurations_from_device_name(
            self, domain: str, instance_id: str):
        """Returns a list of configurations for a given device

        To be implemented in the derived class

        :param domain: DB domain
        :param instance_id: instance id of the device
        :return: a list of dicts:
            [{"configid": uuid of the configuration,
              "instanceid": device instance uuid in the DB},
              ...
            ]
        """
        raise NotImplementedError

    async def get_configurations_from_device_name_part(
            self, domain: str, device_id_part: str):
        """
        Returns a list of configurations for a given device.
        :param domain: DB domain
        :param device_id_part: part of device name; search is case-insensitive.
        :return: a list of dicts:
            [{"config_id": uuid of the configuration,
              "device_uuid": device instance uuid in the DB,
              "device_id": device instance id},
              ...
            ]
        """
        raise NotImplementedError

    async def get_projects_with_conf(self, domain: str, device_id: str):
        """
        Returns a dict with projects and active configurations from a device
        name.

        :param domain: DB domain
        :param device_id: the device to return the information for.
        :return: a dict:
            {"project name": configuration uuid,
             ...}
        """
        raise NotImplementedError

    async def get_projects_from_device(self, domain: str, uuid: str):
        """
        Returns the projects which contain a device instance with a given uuid

        To be implemented in the derived class

        :param domain: DB domain
        :param uuid: the uuid of the device instance from the database
        :return: a set containing project names
        """
        raise NotImplementedError

    async def get_projects_with_device(
            self, domain: str, device_id_part: str) -> list[dict[str, any]]:
        """
        Returns a list of dictionaries with data about projects that contain
        active configurations for a given device.

        :param domain: DB domain
        :param device_id_part: part of name of devices for which project data
                               must be returned.
        :return: a list of dicts:
            [{"projectname": name of project,
              "date": last modification timestamp for the project,
              "uuid": uuid of projecti
              "devices": list of ids of prj devices with the given part}, ...]
        """
        raise NotImplementedError

    async def get_projects_data_from_device(self, domain: str, uuid: str):
        raise NotImplementedError

    async def save_item(
        self, domain: str, uuid: str, item_xml: str,
            overwrite: bool = False):
        raise NotImplementedError

    async def load_item(self, domain: str, items: list[dict]):
        raise NotImplementedError



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from asyncio import ensure_future, gather, sleep

from karabo.native import (
    AccessMode, Bool, Configurable, Hash, Timestamp, UInt32)

from .signalslot import slot

SLOT_PING_SLEEP = 2


class HeartBeatMixin(Configurable):

    track = Bool(
        defaultValue=False,
        displayedName="Track Heartbeats",
        description="This boolean will enable the heart beat tracking",
        accessMode=AccessMode.INITONLY)

    tickInterval = UInt32(
        defaultValue=20,
        minInc=20,
        description="Tick interval to check for heartbeats in seconds",
        accessMode=AccessMode.INITONLY)

    numBeats = UInt32(
        defaultValue=3,
        minInc=2,
        maxInc=10,
        description=("The number of potential beats that are left out before "
                     "an instanceGone is channeled"),
        accessMode=AccessMode.INITONLY)

    def __init__(self, configuration={}):
        super().__init__(configuration)
        self.deviceInstanceMap = {}
        self.systemTopology = Hash("device", Hash(), "server", Hash(),
                                   "macro", Hash(), "client", Hash())

    async def _run(self, **kwargs):
        await super()._run(**kwargs)
        if self.track:
            self.callNoWait("*", "slotDiscover", self.serverId)
            # Before we actually start ticking, we wait a few seconds for all
            # slotInstanceInfos to arrive
            await sleep(SLOT_PING_SLEEP)
            ensure_future(self._sigslot.consume_beats(self))
            ensure_future(self._track_heartbeats())

    @slot
    def slotDiscoverAnswer(self, instanceId, info):
        self._update_instance_info(instanceId, info)

    @slot
    async def slotInstanceNew(self, instanceId, info):
        await super().slotInstanceNew(instanceId, info)
        if self.track:
            await self._add_instance(instanceId, info)

    @slot
    def slotInstanceGone(self, instanceId, info):
        super().slotInstanceGone(instanceId, info)
        if self.track:
            self._remove_instance(instanceId, info)

    @slot
    def slotInstanceUpdated(self, instanceId, info):
        super().slotInstanceUpdated(instanceId, info)
        if self.track:
            self._update_instance_info(instanceId, info)

    async def updateHeartBeat(self, instanceId, info):
        """This method stamps the arrival of a heartbeat notification

        If the ``type`` of the instance is a server it is checked for
        underlying devices and eventually channel a topology event.
        """
        # This method is only called when we are tracking!
        instance_type = info["type"]
        if instanceId not in self.systemTopology[instance_type]:
            # Resurrection, refresh instanceInfo directly
            info = await self._sigslot.request(
                instanceId, "slotPing", 1)
            await self._topology_changed(new=[(instanceId, info)], gone=[])
            await self._add_instance(instanceId, info)
        else:
            self._update_instance_info(instanceId, info)

    # Private interface
    # ----------------------------------------------------------------------

    async def _track_heartbeats(self):
        """Track the heartbeats of all instances"""
        while True:
            await sleep(self.tickInterval)
            await self._track_action()

    async def _track_action(self):
        actual_time = Timestamp()
        instances = []
        for instanceId, _, info in self.systemTopology["device"].iterall():
            if (actual_time - Timestamp.fromHashAttributes(info) >
                    self.numBeats * info["heartbeatInterval"]):
                instances.append((instanceId, info))
        for instanceId, _, info in self.systemTopology["server"].iterall():
            if (actual_time - Timestamp.fromHashAttributes(info) >
                    self.numBeats * info["heartbeatInterval"]):
                instances.append((instanceId, info))

        gone_instances = []
        for instanceId, info in instances:
            gone_instances.extend(self._remove_instance(instanceId, info))
        # Make sure that devices are erased in the beginning
        gone_instances.extend(instances)
        await self._topology_changed(new=[], gone=gone_instances)

    def _update_instance_info(self, instanceId, info):
        """Update the system topology with the instanceInfo from `instance`

        This method adds the timestamp information to the `info`
        """
        info.update(Timestamp().toDict())
        h = Hash()
        h.setElement(
            f"{info['type']}.{instanceId}", Hash(), dict(info.items()))
        self.systemTopology.merge(h)

    def _remove_server_children(
            self, instanceId, info) -> list[tuple[str, dict]]:
        """Cleanup the device children from the server

        Returns: devices if they were still in the topology
        """
        devices = []
        if info["type"] == "server":
            devices = [(k, a) for k, v, a in
                       self.systemTopology["device"].iterall()
                       if a["serverId"] == instanceId]
            for instanceId, _ in devices:
                self.systemTopology["device"].pop(instanceId, None)

        return devices

    async def _add_instance(self, instanceId, info):
        """An instance is freshly added to the topology

        If a server is added to the topology, we check if devices
        are still tracked. Since the server is added, we must
        announce a `slotInstanceGone` for all devices.
        """
        self._update_instance_info(instanceId, info)
        # Check if this was a server coming back online!
        devices = self._remove_server_children(instanceId, info)
        if devices:
            await self._topology_changed(new=[], gone=devices)

    def _remove_instance(self, instanceId, info) -> list[tuple[str, dict]]:
        """An instance is removed from the topology

        Returns a list of tuples (dict, attrs)
        """
        devices = self._remove_server_children(instanceId, info)
        self.systemTopology[info["type"]].pop(instanceId, None)
        return devices

    async def _topology_changed(self, new, gone):
        """Channel the topology tracking information

        This method will send out `slotInstanceGone and `slotInstanceNew`
        to the device children.
        """
        if new:
            await gather(*[dev.slotInstanceNew(instanceId, info)
                           for instanceId, info in new
                           for dev in self.deviceInstanceMap.values()])
        if gone:
            for device in self.deviceInstanceMap.values():
                for instanceId, info in gone:
                    device.slotInstanceGone(instanceId, info)
